import warnings
import logging
import numpy as np
import pytest

from pprint import pprint
from sklearn.tree import DecisionTreeRegressor, ExtraTreeRegressor
from sklearn.metrics.pairwise import rbf_kernel
from sklearn.utils.validation import check_random_state
from sklearn.dummy import DummyRegressor, DummyClassifier
from sklearn.utils._testing import assert_allclose
from sklearn.base import is_classifier

from bipartite_learn.tree import (
    BipartiteDecisionTreeRegressor,
    BipartiteExtraTreeRegressor,
)

from bipartite_learn.ensemble import (
    BipartiteExtraTreesRegressor,
    BipartiteRandomForestRegressor,
)
from sklearn.ensemble import ExtraTreesRegressor, RandomForestRegressor

from bipartite_learn.tree._semisupervised_classes import (
    DecisionTreeRegressorSS,
    BipartiteExtraTreeRegressorSS,
)

from bipartite_learn.melter import (
    row_cartesian_product,
)
from bipartite_learn.ensemble._semisupervised_forest import (
    RandomForestRegressorSS,
    ExtraTreesRegressorSS,
    BipartiteRandomForestRegressorSS,
    BipartiteExtraTreesRegressorSS,
    # BipartiteRandomForestClassifierSS,
)

from .utils.test_utils import (
    gen_mock_data, stopwatch, DEF_PARAMS, parse_args,
)
from .utils.make_examples import (
    make_interaction_blobs,
    make_interaction_regression,
)


DEF_PARAMS = DEF_PARAMS | dict(n_estimators=10)


def eval_model(model, X, y):
    pred = model.predict(X)
    mse = np.mean((pred-y)**2)
    print('* MSE:', mse)
    return mse


def apply_rbf_kernel(regression_data):
    X, Y, x, y = regression_data
    X = [rbf_kernel(X[0]), rbf_kernel(X[1])]
    x = row_cartesian_product(X)
    return X, Y, x, y


@pytest.fixture(params=range(3))
def random_state(request):
    return check_random_state(request.param)


@pytest.fixture
def n_samples(request):
    return (11, 13)


@pytest.fixture#(params=[(50, 30)])
def n_features(request):
    return (10, 10)


@pytest.fixture
def classification_data(n_samples, n_features, random_state):
    X, Y, x, y = make_interaction_blobs(
        return_molten=True,
        n_features=n_features,
        n_samples=n_samples,
        random_state=random_state,
        noise=0.0,
        centers=10,
        row_kwargs={'center_box': (.3, .7), 'cluster_std': .1},
        col_kwargs={'center_box': (.3, .7), 'cluster_std': .1},
    )
    return X, Y, x, y


@pytest.fixture
def regression_data(n_samples, n_features, random_state):
    X, Y, x, y = make_interaction_regression(
        n_samples=n_samples,
        n_features=n_features,
        n_targets=None,
        min_target=0.0,
        max_target=1.0,
        noise=0.0,
        return_molten=True,
        return_tree=False,
        random_state=random_state,
        max_depth=None,
    )
    return X, Y, x, y


@pytest.fixture
def pairwise_regression_data(regression_data):
    return apply_rbf_kernel(regression_data)


@pytest.fixture
def pairwise_classification_data(classification_data):
    return apply_rbf_kernel(classification_data)


@pytest.fixture
def get_appropriate_data(
    classification_data,
    regression_data,
    pairwise_classification_data,
    pairwise_regression_data,
):
    def _getter(estimator):
        if is_classifier(estimator):
            if estimator._get_tags()['pairwise']:
                return pairwise_classification_data
            else:
                return classification_data
        else:
            if estimator._get_tags()['pairwise']:
                return pairwise_regression_data
            else:
                return regression_data
    return _getter


@pytest.mark.parametrize(
    'estimator_name, estimator', [
        (
            'BipartiteRandomForestRegressorSS',
            BipartiteRandomForestRegressorSS(
                unsupervised_criterion_rows='squared_error',
                unsupervised_criterion_cols='squared_error',
            ),
        ),
        (
            'BipartiteRandomForestRegressorSS',
            BipartiteRandomForestRegressorSS(
                unsupervised_criterion_rows='mean_distance',
                unsupervised_criterion_cols='mean_distance',
            ),
        ),
    ],
)
def test_forests_score(
    get_appropriate_data,
    estimator_name,
    estimator,
):
    X, Y, x, y = get_appropriate_data(estimator)

    estimator.fit(X, Y)

    # The two formats must be supported:
    pred1 = estimator.predict(X)
    pred2 = estimator.predict(x)
    assert_allclose(pred1, pred2)

    score = estimator.score(X, y)

    logging.info(f"{estimator_name}'s score: {score}")
    assert score > 0.1


def compare_estimators(estimators1d, estimators2d, **PARAMS):
    PARAMS = DEF_PARAMS | PARAMS

    X, y, X1d, y1d = gen_mock_data(melt=True, **PARAMS)
    y1d = y1d.reshape(-1)

    for name in estimators1d.keys():
        with stopwatch(f'Fitting {name}...'):
            estimators1d[name].fit(X1d, y1d)
    for name in estimators2d.keys():
        with stopwatch(f'Fitting {name}...'):
            estimators2d[name].fit(X, y)

    yvar = y.var()
    print('\nEval baseline, var(y): ', yvar)

    for name, est in (estimators1d|estimators2d).items():
        with stopwatch(f'Evaluating {name}...'):
            assert eval_model(est, X1d, y1d) <= yvar

    return estimators1d, estimators2d


def test_trees(tree_params=None, **PARAMS):
    PARAMS = DEF_PARAMS | PARAMS

    tree_params = tree_params or {}
    tree_params = dict(
        min_samples_leaf=PARAMS['min_samples_leaf'],
        random_state=PARAMS['seed'],
    ) | tree_params

    estimators1d = {
        'DT': DecisionTreeRegressor(**tree_params),
        'sXT': ExtraTreeRegressor(**tree_params),
    }
    estimators2d = {
        'DT2D': BipartiteDecisionTreeRegressor(**tree_params),
        'sXT2D': BipartiteExtraTreeRegressor(**tree_params),
    }
    compare_estimators(estimators1d, estimators2d, **PARAMS)


def test_forests(forest_params=None, **PARAMS):
    PARAMS = DEF_PARAMS | PARAMS

    forest_params = forest_params or {}
    forest_params = dict(
        min_samples_leaf=PARAMS['min_samples_leaf'],
        random_state=PARAMS['seed'],
        n_estimators=PARAMS['n_estimators'],
    ) | forest_params

    estimators1d = {
        'RF': RandomForestRegressor(**forest_params),
        'XT': ExtraTreesRegressor(**forest_params),
    }
    estimators2d = {
        'BRF': BipartiteRandomForestRegressor(**forest_params),
        'BXT': BipartiteExtraTreesRegressor(**forest_params),
    }
    compare_estimators(estimators1d, estimators2d, **PARAMS)


def test_semisupervised_forests(forest_params=None, **PARAMS):
    PARAMS = DEF_PARAMS | PARAMS

    forest_params = forest_params or {}
    forest_params = dict(
        min_samples_leaf=PARAMS['min_samples_leaf'],
        random_state=PARAMS['seed'],
        n_estimators=PARAMS['n_estimators'],
        bootstrap=True,
        max_samples=0.7,
    ) | forest_params
    pprint(forest_params)

    estimators1d = {
        'SSRF': RandomForestRegressorSS(**forest_params),
        'SSXT': ExtraTreesRegressorSS(**forest_params),
        'SSXT2': ExtraTreesRegressorSS(**forest_params),
    }
    estimators2d = {
        'SSBRF': BipartiteRandomForestRegressorSS(**forest_params),
        'SSBXT': BipartiteExtraTreesRegressorSS(**forest_params),
        'SSBXTado': BipartiteExtraTreesRegressorSS(
            axis_decision_only=True,
            **forest_params,
        ),
    }
    compare_estimators(estimators1d, estimators2d, **PARAMS)


@pytest.mark.parametrize(
    'forest_class',
    [
        RandomForestRegressorSS,
        BipartiteRandomForestRegressorSS,
    ],
)
def test_ss_forest_can_set_X_targets(regression_data, forest_class):
    def preprocessor_func(x):
        return x ** 2

    X, Y, x, y = regression_data

    forest = forest_class(
        preprocess_X_targets=preprocessor_func,
        n_estimators=5,
    )

    is_bipartite = forest._get_tags().get('multipartite', False)

    if is_bipartite:
        X_processed = [preprocessor_func(Xi) for Xi in X]
    else:
        X = X[0]
        X_processed = preprocessor_func(X)

    forest.fit(X, Y)

    tree = forest.estimators_[0]
    assert tree.preprocess_X_targets is preprocessor_func
    assert forest._X_targets is not None
    assert tree._X_targets is not None
    assert tree._X_targets is forest._X_targets

    if is_bipartite:
        assert (tree._X_targets[0] == X_processed[0]).all()
        assert (tree._X_targets[1] == X_processed[1]).all()
    else:
        assert (tree._X_targets == X_processed).all()



from .utils.test_utils import gen_mock_data


def test_make_interaction_function(**params):
    params.update(dict(
        shape=(100, 100),
        nattrs=(1, 1),
        nrules=15,
        noise=0,
        seed=0,
        transpose_test=False,
    ))
    X, y = gen_mock_data(**params)

    assert not any(y.all(axis=a).any() for a in range(y.ndim))


# Author: Pedro Ilídio <pedrilidio@gmail.com>
# Modified from scikit-learn.
#
# License: BSD 3 clause

# See _axis_criterion.pyx for implementation details.

from sklearn.tree._criterion cimport Criterion
from sklearn.tree._tree cimport DTYPE_t         # Type of X
from sklearn.tree._tree cimport DOUBLE_t        # Type of y, sample_weight
from sklearn.tree._tree cimport SIZE_t          # Type for indices and counters
from sklearn.tree._tree cimport INT32_t         # Signed 32 bit integer
from sklearn.tree._tree cimport UINT32_t        # Unsigned 32 bit integer


cdef class BaseComposableCriterion(Criterion):
    cdef double _proxy_improvement_factor(self) noexcept nogil


cdef class AxisCriterion(BaseComposableCriterion):
    # AxisCriterion is analogous to sklearn.tree._criterion.Criterion, but can
    # also calculate the impurity on the other axis for the current node and
    # children (as if each child partition were transposed), and has an
    # additional init_columns() method to receive column indices and column
    # weights.

    # When the term sample is used, we are refering to rows. We keep it for
    # compatibility with Criterion objects.

    # y's contiguity is dismissed, since y columns will now be randomly acessed
    cdef const DOUBLE_t[:, :] y_

    # Copied from sklearn.tree._criterion.Criterion
    # cdef const DOUBLE_t[:] sample_weight # Sample weights
    # cdef const SIZE_t[:] sample_indices  # Sample indices in X, y
    # cdef SIZE_t start                    # samples[start:pos] are the samples in the left node
    # cdef SIZE_t pos                      # samples[pos:end] are the samples in the right node
    # cdef SIZE_t end
    # cdef SIZE_t n_outputs                # Number of outputs
    # cdef SIZE_t n_samples                # Number of samples
    # cdef SIZE_t n_node_samples           # Number of samples in the node (end-start)
    # cdef double weighted_n_samples       # Weighted number of samples (in total)
    # cdef double weighted_n_node_samples  # Weighted number of samples in the node
    # cdef double weighted_n_left          # Weighted number of samples in the left node
    # cdef double weighted_n_right         # Weighted number of samples in the right node

    # Attributes specific to AxisCriterion

    cdef const DOUBLE_t[:] col_weights
    cdef const SIZE_t[:] col_indices
    cdef SIZE_t start_col
    cdef SIZE_t end_col

    # Since samples are reordered by splitters in node_split, indices of
    # selected columns must be saved during AxisCriterion.init_columns().
    # Otherwise, Criterion.sum_total will be built in an order that may not
    # correspond to col_indices anymore when the time of calling
    # Criterion.init() comes.

    # In principle, only row_samples needs to be copied, since
    # BipartiteSplitter cals splitter_rows.node_reset() before
    # splitter_cols.node_reset(). However, we choose to copy both
    # to be independent of this specific order.

    # FIXME: If splitters become able of finding the best split without
    # reordering, this strategy can be dropped and we avoid the memory
    # burden.
    cdef SIZE_t[::1] _node_col_indices

    # cdef SIZE_t n_cols  # n_outputs represents the number of columns
    # cdef double weighted_n_cols umn # TODO: not needed for now

    cdef SIZE_t n_node_cols
    cdef double weighted_n_cols
    cdef double weighted_n_node_cols

    cdef bint _columns_are_set

    # FIXME: Since we cannot access rows_impurity and cols_impurity separately,
    # we have to discard self.impurity_improvement()'s input and calculate
    # them again. To mitigate this problem, self.axes_children_impurities()
    # caches the previous values it calculated and reuses them if
    # self.pos == self._cached_pos.
    # NOTE: self._cached_pos must be reset at AxisCriterion.init() to
    # ensure it always corresponds to the current tree node.
    cdef SIZE_t _cached_pos
    cdef double _cached_rows_node_impurity
    cdef double _cached_cols_node_impurity
    cdef double _cached_rows_impurity_left
    cdef double _cached_rows_impurity_right
    cdef double _cached_cols_impurity_left
    cdef double _cached_cols_impurity_right

    cdef void init_columns(
        self,
        const DOUBLE_t[:] col_weights,
        double weighted_n_cols,
        const SIZE_t[:] col_indices,
        SIZE_t start_col,
        SIZE_t end_col,
    ) nogil
    cdef void node_axes_impurities(
        self,
        double* rows_impurity,
        double* cols_impurity,
    ) noexcept nogil
    cdef void children_axes_impurities(
        self,
        double* rows_impurity_left,
        double* rows_impurity_right,
        double* cols_impurity_left,
        double* cols_impurity_right,
    ) noexcept nogil
    cdef int axis_init(
        self,
        const DOUBLE_t[:, :] y,
        const DOUBLE_t[:] sample_weight,
        const DOUBLE_t[:] col_weights,
        const SIZE_t[:] sample_indices,
        const SIZE_t[:] col_indices,
        double weighted_n_samples,
        double weighted_n_cols,
        SIZE_t start,
        SIZE_t end,
        SIZE_t start_col,
        SIZE_t end_col,
    ) except -1 nogil
    cdef void total_node_value(self, double* dest) noexcept nogil
    cdef double _proxy_improvement_factor(self) noexcept nogil


cdef class AxisRegressionCriterion(AxisCriterion):
    """Abstract regression criterion."""

    cdef double sq_sum_total
    # TODO: sq_row_sums is naturally calculated by the criterion in the other
    #       axis. We could set it as a pointer to the other axis criterion's
    #       sum_total.
    cdef double sum_sq_row_sums  # np.sum(sample_weights * np.sum(y, axis=1) ** 2)

    cdef double[::1] sum_total  # The sum of w*y.
    cdef double[::1] sum_left   # Same as above, but for the left side of the split
    cdef double[::1] sum_right  # Same as above, but for the right side of the split


cdef class AxisClassificationCriterion(AxisCriterion):
    """Abstract criterion for classification."""

    cdef SIZE_t[::1] n_classes
    cdef SIZE_t max_n_classes

    cdef double[:, ::1] sum_total   # The sum of the weighted count of each label.
    cdef double[:, ::1] sum_left    # Same as above, but for the left side of the split
    cdef double[:, ::1] sum_right   # Same as above, but for the right side of the split


cdef class AxisCriterionGSO(AxisCriterion):
    cdef DOUBLE_t[::1] _y_row_sums
    cdef double sq_sum_total
    cdef double sum_total  # The sum of w*y.
    cdef double sum_left   # Same as above, but for the left side of the split
    cdef double sum_right  # Same as above, but for the right side of the split


from .loader import (
    get_data_home,
    BaseFileLoader,
    BaseRemoteFileLoader,
    BipartiteDatasetLoader,
)
from ._dpi_yamanishi import (
    EnzymesLoader,
    IonChannelsLoader,
    GPCRLoader,
    NuclearReceptorsLoader,
)

__all__ = [
    "get_data_home",
    "BaseFileLoader",
    "BaseRemoteFileLoader",
    "BipartiteDatasetLoader",
    "EnzymesLoader",
    "IonChannelsLoader",
    "GPCRLoader",
    "NuclearReceptorsLoader",
]


"""
The :mod:`sklearn.model_selection._search` includes utilities to fine-tune the
parameters of an estimator.
"""

# Author: Alexandre Gramfort <alexandre.gramfort@inria.fr>,
#         Gael Varoquaux <gael.varoquaux@normalesup.org>
#         Andreas Mueller <amueller@ais.uni-bonn.de>
#         Olivier Grisel <olivier.grisel@ensta.org>
#         Raghav RV <rvraghav93@gmail.com>
# License: BSD 3 clause

from abc import ABCMeta, abstractmethod
from collections import defaultdict
from functools import partial
from itertools import product
import time
import warnings

import numpy as np
from numpy.ma import MaskedArray
from scipy.stats import rankdata

from sklearn.base import is_classifier, clone
from sklearn.model_selection._validation import _aggregate_score_dicts
from sklearn.model_selection._validation import _insert_error_scores
from sklearn.model_selection._validation import _normalize_score_results
from sklearn.model_selection._validation import _warn_or_raise_about_fit_failures
from sklearn.utils.validation import _check_fit_params
from sklearn.utils.parallel import Parallel, delayed
from sklearn.metrics._scorer import _check_multimetric_scoring
from sklearn.metrics import check_scoring

# Added
from sklearn.model_selection._search import (
    RandomizedSearchCV,
    BaseSearchCV,
    GridSearchCV,
)
from ..base import BaseMultipartiteEstimator
from ._validation import _bipartite_fit_and_score, _check_train_test_combinations
from ._split import check_multipartite_cv

__all__ = ["MultipartiteGridSearchCV", "MultipartiteRandomizedSearchCV"]


class BaseMultipartiteSearchCV(
    BaseMultipartiteEstimator,
    BaseSearchCV,
    metaclass=ABCMeta,
):
    """Abstract base class for hyper parameter search with cross-validation."""

    _parameter_constraints = {
        **BaseSearchCV._parameter_constraints,
        "diagonal": ["boolean"],
        "train_test_combinations": [list, None],
    }

    @abstractmethod
    def __init__(
        self,
        estimator,
        *,
        scoring=None,
        n_jobs=None,
        refit=True,
        cv=None,
        verbose=0,
        pre_dispatch="2*n_jobs",
        error_score=np.nan,
        return_train_score=True,
        # bipartite_learn-specific params:
        diagonal=False,
        train_test_combinations=None,
        pairwise=False,
        **kwargs,
    ):
        self.diagonal = diagonal
        self.train_test_combinations = train_test_combinations
        self.pairwise = pairwise

        super().__init__(
            estimator=estimator,
            scoring=scoring,
            n_jobs=n_jobs,
            refit=refit,
            cv=cv,
            verbose=verbose,
            pre_dispatch=pre_dispatch,
            error_score=error_score,
            return_train_score=return_train_score,
            **kwargs,
        )

    def fit(self, X, y=None, *, groups=None, **fit_params):
        """Run fit with all sets of parameters.

        Parameters
        ----------

        X : array-like of shape (n_samples, n_features)
            Training vector, where `n_samples` is the number of samples and
            `n_features` is the number of features.

        y : array-like of shape (n_samples, n_output) \
            or (n_samples,), default=None
            Target relative to X for classification or regression;
            None for unsupervised learning.

        groups : array-like of shape (n_samples,), default=None
            Group labels for the samples used while splitting the dataset into
            train/test set. Only used in conjunction with a "Group" :term:`cv`
            instance (e.g., :class:`~sklearn.model_selection.GroupKFold`).

        **fit_params : dict of str -> object
            Parameters passed to the `fit` method of the estimator.

            If a fit parameter is an array-like whose length is equal to
            `num_samples` then it will be split across CV groups along with `X`
            and `y`. For example, the :term:`sample_weight` parameter is split
            because `len(sample_weights) = len(X)`.

        Returns
        -------
        self : object
            Instance of fitted estimator.
        """
        estimator = self.estimator
        refit_metric = "score"

        if callable(self.scoring):
            scorers = self.scoring
        elif self.scoring is None or isinstance(self.scoring, str):
            scorers = check_scoring(self.estimator, self.scoring)
        else:
            scorers = _check_multimetric_scoring(self.estimator, self.scoring)
            self._check_refit_for_multimetric(scorers)
            refit_metric = self.refit

        ######### Main changed section
        # FIXME
        # X, y, groups = indexable(X, y, groups)

        fit_params = _check_fit_params(X[0], fit_params)
        cv_orig = check_multipartite_cv(
            self.cv, y, classifier=is_classifier(estimator), diagonal=self.diagonal
        )

        train_test_combinations, train_test_names = _check_train_test_combinations(
            self.train_test_combinations,
            len(X),
            self.return_train_score,
        )
        # This will be used by self._format_results to select the train_score
        self.train_test_names = train_test_names

        fit_and_score_kwargs = dict(
            scorer=scorers,
            fit_params=fit_params,
            return_train_score=self.return_train_score,
            return_n_test_samples=True,
            return_times=True,
            return_parameters=False,
            error_score=self.error_score,
            verbose=self.verbose,
            train_test_combinations=train_test_combinations,
            train_test_names=train_test_names,
            pairwise=self.pairwise,
        )
        ##############################

        n_splits = cv_orig.get_n_splits(X, y, groups)
        base_estimator = clone(self.estimator)
        parallel = Parallel(n_jobs=self.n_jobs, pre_dispatch=self.pre_dispatch)

        results = {}
        with parallel:
            all_candidate_params = []
            all_out = []
            all_more_results = defaultdict(list)

            def evaluate_candidates(candidate_params, cv=None, more_results=None):
                cv = cv or cv_orig
                candidate_params = list(candidate_params)
                n_candidates = len(candidate_params)

                if self.verbose > 0:
                    print(
                        "Fitting {0} folds for each of {1} candidates,"
                        " totalling {2} fits".format(
                            n_splits, n_candidates, n_candidates * n_splits
                        )
                    )

                out = parallel(
                    delayed(_bipartite_fit_and_score)(
                        clone(base_estimator),
                        X,
                        y,
                        train_test=train_test,
                        parameters=parameters,
                        split_progress=(split_idx, n_splits),
                        candidate_progress=(cand_idx, n_candidates),
                        **fit_and_score_kwargs,
                    )
                    for (cand_idx, parameters), (split_idx, train_test) in product(
                        enumerate(candidate_params), enumerate(cv.split(X, y, groups))
                    )
                )

                if len(out) < 1:
                    raise ValueError(
                        "No fits were performed. "
                        "Was the CV iterator empty? "
                        "Were there no candidates?"
                    )
                elif len(out) != n_candidates * n_splits:
                    raise ValueError(
                        "cv.split and cv.get_n_splits returned "
                        "inconsistent results. Expected {} "
                        "splits, got {}".format(n_splits, len(out) // n_candidates)
                    )

                _warn_or_raise_about_fit_failures(out, self.error_score)

                # For callable self.scoring, the return type is only know after
                # calling. If the return type is a dictionary, the error scores
                # can now be inserted with the correct key. The type checking
                # of out will be done in `_insert_error_scores`.
                if callable(self.scoring):
                    _insert_error_scores(out, self.error_score)

                all_candidate_params.extend(candidate_params)
                all_out.extend(out)

                if more_results is not None:
                    for key, value in more_results.items():
                        all_more_results[key].extend(value)

                nonlocal results
                results = self._format_results(
                    all_candidate_params, n_splits, all_out, all_more_results
                )

                return results

            self._run_search(evaluate_candidates)

            # multimetric is determined here because in the case of a callable
            # self.scoring the return type is only known after calling
            first_test_score = all_out[0][train_test_names[-1] + "_scores"]
            self.multimetric_ = isinstance(first_test_score, dict)

            # check refit_metric now for a callabe scorer that is multimetric
            if callable(self.scoring) and self.multimetric_:
                self._check_refit_for_multimetric(first_test_score)
                refit_metric = self.refit

        # For multi-metric evaluation, store the best_index_, best_params_ and
        # best_score_ iff refit is one of the scorer names
        # In single metric evaluation, refit_metric is "score"
        if self.refit or not self.multimetric_:
            self.best_index_ = self._select_best_index(
                self.refit, refit_metric, results
            )
            if not callable(self.refit):
                # With a non-custom callable, we can select the best score
                # based on the best index
                self.best_score_ = results[f"mean_test_{refit_metric}"][
                    self.best_index_
                ]
            self.best_params_ = results["params"][self.best_index_]

        if self.refit:
            # we clone again after setting params in case some
            # of the params are estimators as well.
            self.best_estimator_ = clone(
                clone(base_estimator).set_params(**self.best_params_)
            )
            refit_start_time = time.time()
            if y is not None:
                self.best_estimator_.fit(X, y, **fit_params)
            else:
                self.best_estimator_.fit(X, **fit_params)
            refit_end_time = time.time()
            self.refit_time_ = refit_end_time - refit_start_time

            if hasattr(self.best_estimator_, "feature_names_in_"):
                self.feature_names_in_ = self.best_estimator_.feature_names_in_

        # Store the only scorer not as a dict for single metric evaluation
        self.scorer_ = scorers

        self.cv_results_ = results
        self.n_splits_ = n_splits

        return self

    def _format_results(self, candidate_params, n_splits, out, more_results=None):
        n_candidates = len(candidate_params)
        out = _aggregate_score_dicts(out)

        results = dict(more_results or {})
        for key, val in results.items():
            # each value is a list (as per evaluate_candidate's convention)
            # we convert it to an array for consistency with the other keys
            results[key] = np.asarray(val)

        def _store(key_name, array, weights=None, splits=False, rank=False):
            """A small helper to store the scores/times to the cv_results_"""
            # When iterated first by splits, then by parameters
            # We want `array` to have `n_candidates` rows and `n_splits` cols.
            array = np.array(array, dtype=np.float64).reshape(n_candidates, n_splits)
            if splits:
                for split_idx in range(n_splits):
                    # Uses closure to alter the results
                    results["split%d_%s" % (split_idx, key_name)] = array[:, split_idx]

            array_means = np.average(array, axis=1, weights=weights)
            results["mean_%s" % key_name] = array_means

            # if key_name.startswith(("train_", "test_")) and np.any(
            if key_name.endswith("_scores") and np.any(~np.isfinite(array_means)):
                warnings.warn(
                    f"One or more of the {key_name.split('_')[0]} scores "
                    f"are non-finite: {array_means}",
                    category=UserWarning,
                )

            # Weighted std is not directly available in numpy
            array_stds = np.sqrt(
                np.average(
                    (array - array_means[:, np.newaxis]) ** 2, axis=1, weights=weights
                )
            )
            results["std_%s" % key_name] = array_stds

            if rank:
                results["rank_%s" % key_name] = np.asarray(
                    rankdata(-array_means, method="min"), dtype=np.int32
                )

        _store("fit_time", out["fit_time"])
        _store("score_time", out["score_time"])
        # Use one MaskedArray and mask all the places where the param is not
        # applicable for that candidate. Use defaultdict as each candidate may
        # not contain all the params
        param_results = defaultdict(
            partial(
                MaskedArray,
                np.empty(
                    n_candidates,
                ),
                mask=True,
                dtype=object,
            )
        )
        for cand_idx, params in enumerate(candidate_params):
            for name, value in params.items():
                # An all masked empty array gets created for the key
                # `"param_%s" % name` at the first occurrence of `name`.
                # Setting the value at an index also unmasks that index
                param_results["param_%s" % name][cand_idx] = value

        results.update(param_results)
        # Store a list of param dicts at the key 'params'
        results["params"] = candidate_params

        test_scores_key = self.train_test_names[-1] + "_scores"
        train_scores_key = self.train_test_names[0] + "_scores"

        test_scores_dict = _normalize_score_results(out[test_scores_key])
        if self.return_train_score:
            train_scores_dict = _normalize_score_results(out[train_scores_key])

        for scorer_name in test_scores_dict:
            # Computed the (weighted) mean and std for test scores alone
            _store(
                "test_%s" % scorer_name,
                test_scores_dict[scorer_name],
                splits=True,
                rank=True,
                weights=None,
            )
            if self.return_train_score:
                _store(
                    "train_%s" % scorer_name,
                    train_scores_dict[scorer_name],
                    splits=True,
                )

        return results


class MultipartiteGridSearchCV(BaseMultipartiteSearchCV, GridSearchCV):
    """Exhaustive search over specified parameter values for an estimator.

    Important members are fit, predict.

    GridSearchCV implements a "fit" and a "score" method.
    It also implements "score_samples", "predict", "predict_proba",
    "decision_function", "transform" and "inverse_transform" if they are
    implemented in the estimator used.

    The parameters of the estimator used to apply these methods are optimized
    by cross-validated grid-search over a parameter grid.

    Read more in the :ref:`User Guide <grid_search>`.

    Parameters
    ----------
    estimator : estimator object
        This is assumed to implement the scikit-learn estimator interface.
        Either estimator needs to provide a ``score`` function,
        or ``scoring`` must be passed.

    param_grid : dict or list of dictionaries
        Dictionary with parameters names (`str`) as keys and lists of
        parameter settings to try as values, or a list of such
        dictionaries, in which case the grids spanned by each dictionary
        in the list are explored. This enables searching over any sequence
        of parameter settings.

    scoring : str, callable, list, tuple or dict, default=None
        Strategy to evaluate the performance of the cross-validated model on
        the test set.

        If `scoring` represents a single score, one can use:

        - a single string (see :ref:`scoring_parameter`);
        - a callable (see :ref:`scoring`) that returns a single value.

        If `scoring` represents multiple scores, one can use:

        - a list or tuple of unique strings;
        - a callable returning a dictionary where the keys are the metric
          names and the values are the metric scores;
        - a dictionary with metric names as keys and callables a values.

        See :ref:`multimetric_grid_search` for an example.

    n_jobs : int, default=None
        Number of jobs to run in parallel.
        ``None`` means 1 unless in a :obj:`joblib.parallel_backend` context.
        ``-1`` means using all processors. See :term:`Glossary <n_jobs>`
        for more details.

        .. versionchanged:: v0.20
           `n_jobs` default changed from 1 to None

    refit : bool, str, or callable, default=True
        Refit an estimator using the best found parameters on the whole
        dataset.

        For multiple metric evaluation, this needs to be a `str` denoting the
        scorer that would be used to find the best parameters for refitting
        the estimator at the end.

        Where there are considerations other than maximum score in
        choosing a best estimator, ``refit`` can be set to a function which
        returns the selected ``best_index_`` given ``cv_results_``. In that
        case, the ``best_estimator_`` and ``best_params_`` will be set
        according to the returned ``best_index_`` while the ``best_score_``
        attribute will not be available.

        The refitted estimator is made available at the ``best_estimator_``
        attribute and permits using ``predict`` directly on this
        ``GridSearchCV`` instance.

        Also for multiple metric evaluation, the attributes ``best_index_``,
        ``best_score_`` and ``best_params_`` will only be available if
        ``refit`` is set and all of them will be determined w.r.t this specific
        scorer.

        See ``scoring`` parameter to know more about multiple metric
        evaluation.

        .. versionchanged:: 0.20
            Support for callable added.

    cv : int, cross-validation generator or an iterable, default=None
        Determines the cross-validation splitting strategy.
        Possible inputs for cv are:

        - None, to use the default 5-fold cross validation,
        - integer, to specify the number of folds in a `(Stratified)KFold`,
        - :term:`CV splitter`,
        - An iterable yielding (train, test) splits as arrays of indices.

        For integer/None inputs, if the estimator is a classifier and ``y`` is
        either binary or multiclass, :class:`StratifiedKFold` is used. In all
        other cases, :class:`KFold` is used. These splitters are instantiated
        with `shuffle=False` so the splits will be the same across calls.

        Refer :ref:`User Guide <cross_validation>` for the various
        cross-validation strategies that can be used here.

        .. versionchanged:: 0.22
            ``cv`` default value if None changed from 3-fold to 5-fold.

    verbose : int
        Controls the verbosity: the higher, the more messages.

        - >1 : the computation time for each fold and parameter candidate is
          displayed;
        - >2 : the score is also displayed;
        - >3 : the fold and candidate parameter indexes are also displayed
          together with the starting time of the computation.

    pre_dispatch : int, or str, default='2*n_jobs'
        Controls the number of jobs that get dispatched during parallel
        execution. Reducing this number can be useful to avoid an
        explosion of memory consumption when more jobs get dispatched
        than CPUs can process. This parameter can be:

            - None, in which case all the jobs are immediately
              created and spawned. Use this for lightweight and
              fast-running jobs, to avoid delays due to on-demand
              spawning of the jobs

            - An int, giving the exact number of total jobs that are
              spawned

            - A str, giving an expression as a function of n_jobs,
              as in '2*n_jobs'

    error_score : 'raise' or numeric, default=np.nan
        Value to assign to the score if an error occurs in estimator fitting.
        If set to 'raise', the error is raised. If a numeric value is given,
        FitFailedWarning is raised. This parameter does not affect the refit
        step, which will always raise the error.

    return_train_score : bool, default=False
        If ``False``, the ``cv_results_`` attribute will not include training
        scores.
        Computing training scores is used to get insights on how different
        parameter settings impact the overfitting/underfitting trade-off.
        However computing the scores on the training set can be computationally
        expensive and is not strictly required to select the parameters that
        yield the best generalization performance.

        .. versionadded:: 0.19

        .. versionchanged:: 0.21
            Default value was changed from ``True`` to ``False``

    Attributes
    ----------
    cv_results_ : dict of numpy (masked) ndarrays
        A dict with keys as column headers and values as columns, that can be
        imported into a pandas ``DataFrame``.

        For instance the below given table

        +------------+-----------+------------+-----------------+---+---------+
        |param_kernel|param_gamma|param_degree|split0_test_score|...|rank_t...|
        +============+===========+============+=================+===+=========+
        |  'poly'    |     --    |      2     |       0.80      |...|    2    |
        +------------+-----------+------------+-----------------+---+---------+
        |  'poly'    |     --    |      3     |       0.70      |...|    4    |
        +------------+-----------+------------+-----------------+---+---------+
        |  'rbf'     |     0.1   |     --     |       0.80      |...|    3    |
        +------------+-----------+------------+-----------------+---+---------+
        |  'rbf'     |     0.2   |     --     |       0.93      |...|    1    |
        +------------+-----------+------------+-----------------+---+---------+

        will be represented by a ``cv_results_`` dict of::

            {
            'param_kernel': masked_array(data = ['poly', 'poly', 'rbf', 'rbf'],
                                         mask = [False False False False]...)
            'param_gamma': masked_array(data = [-- -- 0.1 0.2],
                                        mask = [ True  True False False]...),
            'param_degree': masked_array(data = [2.0 3.0 -- --],
                                         mask = [False False  True  True]...),
            'split0_test_score'  : [0.80, 0.70, 0.80, 0.93],
            'split1_test_score'  : [0.82, 0.50, 0.70, 0.78],
            'mean_test_score'    : [0.81, 0.60, 0.75, 0.85],
            'std_test_score'     : [0.01, 0.10, 0.05, 0.08],
            'rank_test_score'    : [2, 4, 3, 1],
            'split0_train_score' : [0.80, 0.92, 0.70, 0.93],
            'split1_train_score' : [0.82, 0.55, 0.70, 0.87],
            'mean_train_score'   : [0.81, 0.74, 0.70, 0.90],
            'std_train_score'    : [0.01, 0.19, 0.00, 0.03],
            'mean_fit_time'      : [0.73, 0.63, 0.43, 0.49],
            'std_fit_time'       : [0.01, 0.02, 0.01, 0.01],
            'mean_score_time'    : [0.01, 0.06, 0.04, 0.04],
            'std_score_time'     : [0.00, 0.00, 0.00, 0.01],
            'params'             : [{'kernel': 'poly', 'degree': 2}, ...],
            }

        NOTE

        The key ``'params'`` is used to store a list of parameter
        settings dicts for all the parameter candidates.

        The ``mean_fit_time``, ``std_fit_time``, ``mean_score_time`` and
        ``std_score_time`` are all in seconds.

        For multi-metric evaluation, the scores for all the scorers are
        available in the ``cv_results_`` dict at the keys ending with that
        scorer's name (``'_<scorer_name>'``) instead of ``'_score'`` shown
        above. ('split0_test_precision', 'mean_train_precision' etc.)

    best_estimator_ : estimator
        Estimator that was chosen by the search, i.e. estimator
        which gave highest score (or smallest loss if specified)
        on the left out data. Not available if ``refit=False``.

        See ``refit`` parameter for more information on allowed values.

    best_score_ : float
        Mean cross-validated score of the best_estimator

        For multi-metric evaluation, this is present only if ``refit`` is
        specified.

        This attribute is not available if ``refit`` is a function.

    best_params_ : dict
        Parameter setting that gave the best results on the hold out data.

        For multi-metric evaluation, this is present only if ``refit`` is
        specified.

    best_index_ : int
        The index (of the ``cv_results_`` arrays) which corresponds to the best
        candidate parameter setting.

        The dict at ``search.cv_results_['params'][search.best_index_]`` gives
        the parameter setting for the best model, that gives the highest
        mean score (``search.best_score_``).

        For multi-metric evaluation, this is present only if ``refit`` is
        specified.

    scorer_ : function or a dict
        Scorer function used on the held out data to choose the best
        parameters for the model.

        For multi-metric evaluation, this attribute holds the validated
        ``scoring`` dict which maps the scorer key to the scorer callable.

    n_splits_ : int
        The number of cross-validation splits (folds/iterations).

    refit_time_ : float
        Seconds used for refitting the best model on the whole dataset.

        This is present only if ``refit`` is not False.

        .. versionadded:: 0.20

    multimetric_ : bool
        Whether or not the scorers compute several metrics.

    classes_ : ndarray of shape (n_classes,)
        The classes labels. This is present only if ``refit`` is specified and
        the underlying estimator is a classifier.

    n_features_in_ : int
        Number of features seen during :term:`fit`. Only defined if
        `best_estimator_` is defined (see the documentation for the `refit`
        parameter for more details) and that `best_estimator_` exposes
        `n_features_in_` when fit.

        .. versionadded:: 0.24

    feature_names_in_ : ndarray of shape (`n_features_in_`,)
        Names of features seen during :term:`fit`. Only defined if
        `best_estimator_` is defined (see the documentation for the `refit`
        parameter for more details) and that `best_estimator_` exposes
        `feature_names_in_` when fit.

        .. versionadded:: 1.0

    See Also
    --------
    ParameterGrid : Generates all the combinations of a hyperparameter grid.
    train_test_split : Utility function to split the data into a development
        set usable for fitting a GridSearchCV instance and an evaluation set
        for its final evaluation.
    sklearn.metrics.make_scorer : Make a scorer from a performance metric or
        loss function.

    Notes
    -----
    The parameters selected are those that maximize the score of the left out
    data, unless an explicit score is passed in which case it is used instead.

    If `n_jobs` was set to a value higher than one, the data is copied for each
    point in the grid (and not `n_jobs` times). This is done for efficiency
    reasons if individual jobs take very little time, but may raise errors if
    the dataset is large and not enough memory is available.  A workaround in
    this case is to set `pre_dispatch`. Then, the memory is copied only
    `pre_dispatch` many times. A reasonable value for `pre_dispatch` is `2 *
    n_jobs`.

    Examples
    --------
    >>> from sklearn import svm, datasets
    >>> from sklearn.model_selection import GridSearchCV
    >>> iris = datasets.load_iris()
    >>> parameters = {'kernel':('linear', 'rbf'), 'C':[1, 10]}
    >>> svc = svm.SVC()
    >>> clf = GridSearchCV(svc, parameters)
    >>> clf.fit(iris.data, iris.target)
    GridSearchCV(estimator=SVC(),
                 param_grid={'C': [1, 10], 'kernel': ('linear', 'rbf')})
    >>> sorted(clf.cv_results_.keys())
    ['mean_fit_time', 'mean_score_time', 'mean_test_score',...
     'param_C', 'param_kernel', 'params',...
     'rank_test_score', 'split0_test_score',...
     'split2_test_score', ...
     'std_fit_time', 'std_score_time', 'std_test_score']
    """

    _required_parameters = ["estimator", "param_grid"]

    def __init__(
        self,
        estimator,
        param_grid,
        *,
        scoring=None,
        n_jobs=None,
        refit=True,
        cv=None,
        verbose=0,
        pre_dispatch="2*n_jobs",
        error_score=np.nan,
        return_train_score=False,
        # bipartite_learn-specific params:
        diagonal=False,
        train_test_combinations=None,
        pairwise=False,
    ):
        super().__init__(
            estimator=estimator,
            param_grid=param_grid,
            scoring=scoring,
            n_jobs=n_jobs,
            refit=refit,
            cv=cv,
            verbose=verbose,
            pre_dispatch=pre_dispatch,
            error_score=error_score,
            return_train_score=return_train_score,
            diagonal=diagonal,
            train_test_combinations=train_test_combinations,
            pairwise=pairwise,
        )


class MultipartiteRandomizedSearchCV(
    BaseMultipartiteSearchCV,
    RandomizedSearchCV,
):
    """Randomized search on hyper parameters.

    RandomizedSearchCV implements a "fit" and a "score" method.
    It also implements "score_samples", "predict", "predict_proba",
    "decision_function", "transform" and "inverse_transform" if they are
    implemented in the estimator used.

    The parameters of the estimator used to apply these methods are optimized
    by cross-validated search over parameter settings.

    In contrast to GridSearchCV, not all parameter values are tried out, but
    rather a fixed number of parameter settings is sampled from the specified
    distributions. The number of parameter settings that are tried is
    given by n_iter.

    If all parameters are presented as a list,
    sampling without replacement is performed. If at least one parameter
    is given as a distribution, sampling with replacement is used.
    It is highly recommended to use continuous distributions for continuous
    parameters.

    Read more in the :ref:`User Guide <randomized_parameter_search>`.

    .. versionadded:: 0.14

    Parameters
    ----------
    estimator : estimator object
        A object of that type is instantiated for each grid point.
        This is assumed to implement the scikit-learn estimator interface.
        Either estimator needs to provide a ``score`` function,
        or ``scoring`` must be passed.

    param_distributions : dict or list of dicts
        Dictionary with parameters names (`str`) as keys and distributions
        or lists of parameters to try. Distributions must provide a ``rvs``
        method for sampling (such as those from scipy.stats.distributions).
        If a list is given, it is sampled uniformly.
        If a list of dicts is given, first a dict is sampled uniformly, and
        then a parameter is sampled using that dict as above.

    n_iter : int, default=10
        Number of parameter settings that are sampled. n_iter trades
        off runtime vs quality of the solution.

    scoring : str, callable, list, tuple or dict, default=None
        Strategy to evaluate the performance of the cross-validated model on
        the test set.

        If `scoring` represents a single score, one can use:

        - a single string (see :ref:`scoring_parameter`);
        - a callable (see :ref:`scoring`) that returns a single value.

        If `scoring` represents multiple scores, one can use:

        - a list or tuple of unique strings;
        - a callable returning a dictionary where the keys are the metric
          names and the values are the metric scores;
        - a dictionary with metric names as keys and callables a values.

        See :ref:`multimetric_grid_search` for an example.

        If None, the estimator's score method is used.

    n_jobs : int, default=None
        Number of jobs to run in parallel.
        ``None`` means 1 unless in a :obj:`joblib.parallel_backend` context.
        ``-1`` means using all processors. See :term:`Glossary <n_jobs>`
        for more details.

        .. versionchanged:: v0.20
           `n_jobs` default changed from 1 to None

    refit : bool, str, or callable, default=True
        Refit an estimator using the best found parameters on the whole
        dataset.

        For multiple metric evaluation, this needs to be a `str` denoting the
        scorer that would be used to find the best parameters for refitting
        the estimator at the end.

        Where there are considerations other than maximum score in
        choosing a best estimator, ``refit`` can be set to a function which
        returns the selected ``best_index_`` given the ``cv_results``. In that
        case, the ``best_estimator_`` and ``best_params_`` will be set
        according to the returned ``best_index_`` while the ``best_score_``
        attribute will not be available.

        The refitted estimator is made available at the ``best_estimator_``
        attribute and permits using ``predict`` directly on this
        ``RandomizedSearchCV`` instance.

        Also for multiple metric evaluation, the attributes ``best_index_``,
        ``best_score_`` and ``best_params_`` will only be available if
        ``refit`` is set and all of them will be determined w.r.t this specific
        scorer.

        See ``scoring`` parameter to know more about multiple metric
        evaluation.

        .. versionchanged:: 0.20
            Support for callable added.

    cv : int, cross-validation generator or an iterable, default=None
        Determines the cross-validation splitting strategy.
        Possible inputs for cv are:

        - None, to use the default 5-fold cross validation,
        - integer, to specify the number of folds in a `(Stratified)KFold`,
        - :term:`CV splitter`,
        - An iterable yielding (train, test) splits as arrays of indices.

        For integer/None inputs, if the estimator is a classifier and ``y`` is
        either binary or multiclass, :class:`StratifiedKFold` is used. In all
        other cases, :class:`KFold` is used. These splitters are instantiated
        with `shuffle=False` so the splits will be the same across calls.

        Refer :ref:`User Guide <cross_validation>` for the various
        cross-validation strategies that can be used here.

        .. versionchanged:: 0.22
            ``cv`` default value if None changed from 3-fold to 5-fold.

    verbose : int
        Controls the verbosity: the higher, the more messages.

    pre_dispatch : int, or str, default='2*n_jobs'
        Controls the number of jobs that get dispatched during parallel
        execution. Reducing this number can be useful to avoid an
        explosion of memory consumption when more jobs get dispatched
        than CPUs can process. This parameter can be:

            - None, in which case all the jobs are immediately
              created and spawned. Use this for lightweight and
              fast-running jobs, to avoid delays due to on-demand
              spawning of the jobs

            - An int, giving the exact number of total jobs that are
              spawned

            - A str, giving an expression as a function of n_jobs,
              as in '2*n_jobs'

    random_state : int, RandomState instance or None, default=None
        Pseudo random number generator state used for random uniform sampling
        from lists of possible values instead of scipy.stats distributions.
        Pass an int for reproducible output across multiple
        function calls.
        See :term:`Glossary <random_state>`.

    error_score : 'raise' or numeric, default=np.nan
        Value to assign to the score if an error occurs in estimator fitting.
        If set to 'raise', the error is raised. If a numeric value is given,
        FitFailedWarning is raised. This parameter does not affect the refit
        step, which will always raise the error.

    return_train_score : bool, default=False
        If ``False``, the ``cv_results_`` attribute will not include training
        scores.
        Computing training scores is used to get insights on how different
        parameter settings impact the overfitting/underfitting trade-off.
        However computing the scores on the training set can be computationally
        expensive and is not strictly required to select the parameters that
        yield the best generalization performance.

        .. versionadded:: 0.19

        .. versionchanged:: 0.21
            Default value was changed from ``True`` to ``False``

    Attributes
    ----------
    cv_results_ : dict of numpy (masked) ndarrays
        A dict with keys as column headers and values as columns, that can be
        imported into a pandas ``DataFrame``.

        For instance the below given table

        +--------------+-------------+-------------------+---+---------------+
        | param_kernel | param_gamma | split0_test_score |...|rank_test_score|
        +==============+=============+===================+===+===============+
        |    'rbf'     |     0.1     |       0.80        |...|       1       |
        +--------------+-------------+-------------------+---+---------------+
        |    'rbf'     |     0.2     |       0.84        |...|       3       |
        +--------------+-------------+-------------------+---+---------------+
        |    'rbf'     |     0.3     |       0.70        |...|       2       |
        +--------------+-------------+-------------------+---+---------------+

        will be represented by a ``cv_results_`` dict of::

            {
            'param_kernel' : masked_array(data = ['rbf', 'rbf', 'rbf'],
                                          mask = False),
            'param_gamma'  : masked_array(data = [0.1 0.2 0.3], mask = False),
            'split0_test_score'  : [0.80, 0.84, 0.70],
            'split1_test_score'  : [0.82, 0.50, 0.70],
            'mean_test_score'    : [0.81, 0.67, 0.70],
            'std_test_score'     : [0.01, 0.24, 0.00],
            'rank_test_score'    : [1, 3, 2],
            'split0_train_score' : [0.80, 0.92, 0.70],
            'split1_train_score' : [0.82, 0.55, 0.70],
            'mean_train_score'   : [0.81, 0.74, 0.70],
            'std_train_score'    : [0.01, 0.19, 0.00],
            'mean_fit_time'      : [0.73, 0.63, 0.43],
            'std_fit_time'       : [0.01, 0.02, 0.01],
            'mean_score_time'    : [0.01, 0.06, 0.04],
            'std_score_time'     : [0.00, 0.00, 0.00],
            'params'             : [{'kernel' : 'rbf', 'gamma' : 0.1}, ...],
            }

        NOTE

        The key ``'params'`` is used to store a list of parameter
        settings dicts for all the parameter candidates.

        The ``mean_fit_time``, ``std_fit_time``, ``mean_score_time`` and
        ``std_score_time`` are all in seconds.

        For multi-metric evaluation, the scores for all the scorers are
        available in the ``cv_results_`` dict at the keys ending with that
        scorer's name (``'_<scorer_name>'``) instead of ``'_score'`` shown
        above. ('split0_test_precision', 'mean_train_precision' etc.)

    best_estimator_ : estimator
        Estimator that was chosen by the search, i.e. estimator
        which gave highest score (or smallest loss if specified)
        on the left out data. Not available if ``refit=False``.

        For multi-metric evaluation, this attribute is present only if
        ``refit`` is specified.

        See ``refit`` parameter for more information on allowed values.

    best_score_ : float
        Mean cross-validated score of the best_estimator.

        For multi-metric evaluation, this is not available if ``refit`` is
        ``False``. See ``refit`` parameter for more information.

        This attribute is not available if ``refit`` is a function.

    best_params_ : dict
        Parameter setting that gave the best results on the hold out data.

        For multi-metric evaluation, this is not available if ``refit`` is
        ``False``. See ``refit`` parameter for more information.

    best_index_ : int
        The index (of the ``cv_results_`` arrays) which corresponds to the best
        candidate parameter setting.

        The dict at ``search.cv_results_['params'][search.best_index_]`` gives
        the parameter setting for the best model, that gives the highest
        mean score (``search.best_score_``).

        For multi-metric evaluation, this is not available if ``refit`` is
        ``False``. See ``refit`` parameter for more information.

    scorer_ : function or a dict
        Scorer function used on the held out data to choose the best
        parameters for the model.

        For multi-metric evaluation, this attribute holds the validated
        ``scoring`` dict which maps the scorer key to the scorer callable.

    n_splits_ : int
        The number of cross-validation splits (folds/iterations).

    refit_time_ : float
        Seconds used for refitting the best model on the whole dataset.

        This is present only if ``refit`` is not False.

        .. versionadded:: 0.20

    multimetric_ : bool
        Whether or not the scorers compute several metrics.

    classes_ : ndarray of shape (n_classes,)
        The classes labels. This is present only if ``refit`` is specified and
        the underlying estimator is a classifier.

    n_features_in_ : int
        Number of features seen during :term:`fit`. Only defined if
        `best_estimator_` is defined (see the documentation for the `refit`
        parameter for more details) and that `best_estimator_` exposes
        `n_features_in_` when fit.

        .. versionadded:: 0.24

    feature_names_in_ : ndarray of shape (`n_features_in_`,)
        Names of features seen during :term:`fit`. Only defined if
        `best_estimator_` is defined (see the documentation for the `refit`
        parameter for more details) and that `best_estimator_` exposes
        `feature_names_in_` when fit.

        .. versionadded:: 1.0

    See Also
    --------
    GridSearchCV : Does exhaustive search over a grid of parameters.
    ParameterSampler : A generator over parameter settings, constructed from
        param_distributions.

    Notes
    -----
    The parameters selected are those that maximize the score of the held-out
    data, according to the scoring parameter.

    If `n_jobs` was set to a value higher than one, the data is copied for each
    parameter setting(and not `n_jobs` times). This is done for efficiency
    reasons if individual jobs take very little time, but may raise errors if
    the dataset is large and not enough memory is available.  A workaround in
    this case is to set `pre_dispatch`. Then, the memory is copied only
    `pre_dispatch` many times. A reasonable value for `pre_dispatch` is `2 *
    n_jobs`.

    Examples
    --------
    >>> from sklearn.datasets import load_iris
    >>> from sklearn.linear_model import LogisticRegression
    >>> from sklearn.model_selection import RandomizedSearchCV
    >>> from scipy.stats import uniform
    >>> iris = load_iris()
    >>> logistic = LogisticRegression(solver='saga', tol=1e-2, max_iter=200,
    ...                               random_state=0)
    >>> distributions = dict(C=uniform(loc=0, scale=4),
    ...                      penalty=['l2', 'l1'])
    >>> clf = RandomizedSearchCV(logistic, distributions, random_state=0)
    >>> search = clf.fit(iris.data, iris.target)
    >>> search.best_params_
    {'C': 2..., 'penalty': 'l1'}
    """

    _required_parameters = ["estimator", "param_distributions"]

    def __init__(
        self,
        estimator,
        param_distributions,
        *,
        n_iter=10,
        scoring=None,
        n_jobs=None,
        refit=True,
        cv=None,
        verbose=0,
        pre_dispatch="2*n_jobs",
        random_state=None,
        error_score=np.nan,
        return_train_score=False,
        # bipartite_learn-specific params:
        diagonal=False,
        train_test_combinations=None,
        pairwise=True,
    ):
        super().__init__(
            estimator=estimator,
            param_distributions=param_distributions,
            n_iter=n_iter,
            scoring=scoring,
            n_jobs=n_jobs,
            refit=refit,
            cv=cv,
            verbose=verbose,
            pre_dispatch=pre_dispatch,
            random_state=random_state,
            error_score=error_score,
            return_train_score=return_train_score,
            diagonal=diagonal,
            train_test_combinations=train_test_combinations,
            pairwise=pairwise,
        )



"""Drug-protein interaction prediction datasets from Yamanishi et al., 2008.

The original files can be downloaded from:
    http://web.kuicr.kyoto-u.ac.jp/supp/yoshi/drugtarget/

This module provides utilities for fetching and loading the four gold-standard
datasets provided by [1]_. Each dataset is comprised of two feature matrices
describing a set of drugs and a set of target proteins, as well as a binary
interaction matrix determining the interaction between a drug and a protein is
confirmed (value 1) or that no interaction is known to exist between them (value
0).

The four datasets correspond to four protein families: enzymes, ion channels,
GPCRs and nuclear receptors. The input features are similarity matrices between
the instances on each axis. The score of a Smith-Waterman pairwise alignment is
taken as the the similarity between proteins, whereas the SIMCOMP score is used
for the similarity between drug molecules.

References
----------
.. [1] Yoshihiro Yamanishi, Michihiro Araki, Alex Gutteridge, Wataru Honda,
Minoru Kanehisa, Prediction of drug–target interaction networks from the
integration of chemical and genomic spaces, Bioinformatics, Volume 24, Issue 13,
July 2008, Pages i232–i240, https://doi.org/10.1093/bioinformatics/btn162
"""
# Author: Pedro Ilídio <ilidio@alumni.usp.br>
# License: BSD 3 clause

from pathlib import Path
from sklearn.utils import check_pandas_support, check_symmetric
from .loader import get_data_home, BaseRemoteFileLoader, BipartiteDatasetLoader

__all__ = [
    "EnzymesLoader",
    "IonChannelsLoader",
    "GPCRsLoader",
    "NuclearReceptorsLoader",
]

# The original data can be found here.
BASE_URL = "http://web.kuicr.kyoto-u.ac.jp/supp/yoshi/drugtarget/"


class YamanishiFileLoader(BaseRemoteFileLoader):
    def load_local(self, as_frame=False):
        check_pandas_support(f"{type(self).__name__}.load_local()")
        import pandas as pd
        data = pd.read_table(self.local_path, index_col=0)
        if as_frame:
            return data
        return data.values


class EnzymesLoader(BipartiteDatasetLoader):
    """Binary interaction prediction between enzymes and drug molecules.

    This is one of four gold-standand datasets for drug-protein interaction
    prediction introduced by Yamanishi et al., 2008 [1]_.
    
    The input features are similarity matrices among the instances on each
    axis and the target matrix is a binary interaction matrix determining the
    the existence of an experimentally validated interaction (value 1) or the
    absence of information about an interaction (value 0).

    The score of a Smith-Waterman pairwise alignment is taken as the the
    similarity between proteins, whereas the SIMCOMP score is used for the
    similarity between drug molecules.

    The original files can be downloaded from:
        http://web.kuicr.kyoto-u.ac.jp/supp/yoshi/drugtarget/

    References
    ----------
    .. [1] Yoshihiro Yamanishi, Michihiro Araki, Alex Gutteridge, Wataru Honda,
    Minoru Kanehisa, Prediction of drug–target interaction networks from the
    integration of chemical and genomic spaces, Bioinformatics, Volume 24, Issue
    13, July 2008, Pages i232–i240,
    https://doi.org/10.1093/bioinformatics/btn162
    """

    def __init__(self, base_dir: str | Path | None = None):
        """Initialize the loader.
        Parameters
        ----------
        base_dir : str | Path | None, default=None
            The path to the directory where the data will be stored. If None,
            the data will be stored in the directory returned by
            :func:`bipartite_learn.datasets.get_data_home`.
        """
        super().__init__(
            filepath="dpi_enzymes",
            base_dir=base_dir or get_data_home(),
            y_loader=YamanishiFileLoader(
                url=BASE_URL + "e_admat_dgc.txt",
                checksum=(
                    "bf73f8f0c1a71f9fe92d1e0d8a1065a1122b383532cd47fc73df1cb0d"
                    "9ba6c10"
                ),
            ),
            X_loader=[
                YamanishiFileLoader(
                    url=BASE_URL + "e_simmat_dg.txt",
                    checksum=(
                        "14816c2bba2193d62f52ab53bd51fe9a490b28e98cf33df392a5d"
                        "a2bf24e5607"
                    ),
                ),
                YamanishiFileLoader(
                    url=BASE_URL + "e_simmat_dc.txt",
                    checksum=(
                        "28105b9570c91b4eac688eaf92db2e4aacb7117bc8665cab527fc"
                        "5d45cd83606"
                    ),
                ),
            ],
        )


class IonChannelsLoader(BipartiteDatasetLoader):
    """Binary interactions between ion channels and drug molecules.

    This is one of four gold-standand datasets for drug-protein interaction
    prediction introduced by Yamanishi et al., 2008 [1]_.
    
    The input features are similarity matrices among the instances on each
    axis and the target matrix is a binary interaction matrix determining the
    the existence of an experimentally validated interaction (value 1) or the
    absence of information about an interaction (value 0).

    The score of a Smith-Waterman pairwise alignment is taken as the the
    similarity between proteins, whereas the SIMCOMP score is used for the
    similarity between drug molecules.

    The original files can be downloaded from:
        http://web.kuicr.kyoto-u.ac.jp/supp/yoshi/drugtarget/

    References
    ----------
    .. [1] Yoshihiro Yamanishi, Michihiro Araki, Alex Gutteridge, Wataru Honda,
    Minoru Kanehisa, Prediction of drug–target interaction networks from the
    integration of chemical and genomic spaces, Bioinformatics, Volume 24, Issue
    13, July 2008, Pages i232–i240,
    https://doi.org/10.1093/bioinformatics/btn162
    """

    def __init__(self, base_dir: str | Path | None = None):
        """Initialize the loader.
        Parameters
        ----------
        base_dir : str | Path | None, default=None
            The path to the directory where the data will be stored. If None,
            the data will be stored in the directory returned by
            :func:`bipartite_learn.datasets.get_data_home`.
        """
        super().__init__(
            filepath="dpi_ion_channels",
            base_dir=base_dir or get_data_home(),
            y_loader=YamanishiFileLoader(
                url=BASE_URL + "ic_admat_dgc.txt",
                checksum=(
                    "ac9df3498bc12af5174503e5f1e09ce26a6d78d753861a5eca55425f9"
                    "f9ae2c9"
                ),
            ),
            X_loader=[
                YamanishiFileLoader(
                    url=BASE_URL + "ic_simmat_dg.txt",
                    checksum=(
                        "e15626145623124ad42a45412c544d5fed5e079003727df784d29"
                        "ed3ca72efef"
                    ),
                ),
                YamanishiFileLoader(
                    url=BASE_URL + "ic_simmat_dc.txt",
                    checksum=(
                        "c1078ca88528d75ddb43d4ed2559391d7d8f5375ec572e6e9112e"
                        "519d9cdbbf9"
                    ),
                ),
            ],
        )


class GPCRLoader(BipartiteDatasetLoader):
    """Binary interactions between G-protein coupled receptors and drug molecules.

    This is one of four gold-standand datasets for drug-protein interaction
    prediction introduced by Yamanishi et al., 2008 [1]_.
    
    The input features are similarity matrices among the instances on each
    axis and the target matrix is a binary interaction matrix determining the
    the existence of an experimentally validated interaction (value 1) or the
    absence of information about an interaction (value 0).

    The score of a Smith-Waterman pairwise alignment is taken as the the
    similarity between proteins, whereas the SIMCOMP score is used for the
    similarity between drug molecules.

    The original files can be downloaded from:
        http://web.kuicr.kyoto-u.ac.jp/supp/yoshi/drugtarget/

    References
    ----------
    .. [1] Yoshihiro Yamanishi, Michihiro Araki, Alex Gutteridge, Wataru Honda,
    Minoru Kanehisa, Prediction of drug–target interaction networks from the
    integration of chemical and genomic spaces, Bioinformatics, Volume 24, Issue
    13, July 2008, Pages i232–i240,
    https://doi.org/10.1093/bioinformatics/btn162
    """

    def __init__(self, base_dir: str | Path | None = None):
        """Initialize the loader.
        Parameters
        ----------
        base_dir : str | Path | None, default=None
            The path to the directory where the data will be stored. If None,
            the data will be stored in the directory returned by
            :func:`bipartite_learn.datasets.get_data_home`.
        """
        super().__init__(
            filepath="dpi_gpcr",
            base_dir=base_dir or get_data_home(),
            y_loader=YamanishiFileLoader(
                url=BASE_URL + "gpcr_admat_dgc.txt",
                checksum=(
                    "49b2a8a0139dac3116399618bfefc9d1afcc276ce31b7de4e241d1b6e"
                    "62ef8ab"
                ),
            ),
            X_loader=[
                YamanishiFileLoader(
                    url=BASE_URL + "gpcr_simmat_dg.txt",
                    checksum=(
                        "c58c4311d2086a6914208ed0914598528ce54ade88eb52b0f737e"
                        "edcaf5001e4"
                    ),
                ),
                YamanishiFileLoader(
                    url=BASE_URL + "gpcr_simmat_dc.txt",
                    checksum=(
                        "ba9c1cf096f0f688d1341fec22ebdecb63162005394d3b5ec3e2f"
                        "2ef3fd40df9"
                    ),
                ),
            ],
        )


class NuclearReceptorsLoader(BipartiteDatasetLoader):
    """Binary interactions between nuclear receptors and drug molecules.

    This is one of four gold-standand datasets for drug-protein interaction
    prediction introduced by Yamanishi et al., 2008 [1]_.
    
    The input features are similarity matrices among the instances on each
    axis and the target matrix is a binary interaction matrix determining the
    the existence of an experimentally validated interaction (value 1) or the
    absence of information about an interaction (value 0).

    The score of a Smith-Waterman pairwise alignment is taken as the the
    similarity between proteins, whereas the SIMCOMP score is used for the
    similarity between drug molecules.

    The original files can be downloaded from:
        http://web.kuicr.kyoto-u.ac.jp/supp/yoshi/drugtarget/

    References
    ----------
    .. [1] Yoshihiro Yamanishi, Michihiro Araki, Alex Gutteridge, Wataru Honda,
    Minoru Kanehisa, Prediction of drug–target interaction networks from the
    integration of chemical and genomic spaces, Bioinformatics, Volume 24, Issue
    13, July 2008, Pages i232–i240,
    https://doi.org/10.1093/bioinformatics/btn162
    """

    def __init__(self, base_dir: str | Path | None = None):
        """Initialize the loader.
        Parameters
        ----------
        base_dir : str | Path | None, default=None
            The path to the directory where the data will be stored. If None,
            the data will be stored in the directory returned by
            :func:`bipartite_learn.datasets.get_data_home`.
        """
        super().__init__(
            filepath="dpi_nuclear_receptors",
            base_dir=base_dir or get_data_home(),
            y_loader=YamanishiFileLoader(
                url=BASE_URL + "nr_admat_dgc.txt",
                checksum=(
                    "74256ec88b415dbdb93735e93797955e82587f3154988e99787a4e604"
                    "111963b"
                ),
            ),
            X_loader=[
                YamanishiFileLoader(
                    url=BASE_URL + "nr_simmat_dg.txt",
                    checksum=(
                        "5bf651f310b037320fe5d8aa61dfa15e0d8af4b0fd50c39692010"
                        "85e64b631e1"
                    ),
                ),
                YamanishiFileLoader(
                    url=BASE_URL + "nr_simmat_dc.txt",
                    checksum=(
                        "f80e101f284934b85c5dc7601ecaca2cc86c62a946141b2587f3a"
                        "c8482c9a7c1"
                    ),
                ),
            ],
        )


import logging
from pathlib import Path
from time import time
import numpy as np
from functools import partial
import pytest

from sklearn.tree import DecisionTreeRegressor
from sklearn.tree._splitter import BestSplitter
from sklearn.utils import check_random_state

from bipartite_learn.tree import BipartiteDecisionTreeRegressor
from bipartite_learn.tree._splitter_factory import (
    make_bipartite_ss_splitter,
    make_semisupervised_criterion,
) 
from bipartite_learn.tree._semisupervised_criterion import (
    SSCompositeCriterion,
)

from bipartite_learn.tree._semisupervised_classes import (
    DecisionTreeRegressorSS, BipartiteDecisionTreeRegressorSS,
)

from sklearn.tree._criterion import MSE
from .test_bipartite_trees import compare_trees, parse_args

# from sklearn.tree._tree import DTYPE_t, DOUBLE_t
DTYPE_t, DOUBLE_t = np.float32, np.float64

logging.basicConfig(level=logging.DEBUG)
logging.getLogger("matplotlib").setLevel(logging.CRITICAL)

# Default test params
DEF_PARAMS = dict(
    # n_samples=(150, 160),
    n_samples=(50, 60),
    n_features=(10, 9),
    # n_targets=(2, 1),
    min_target=0.0,
    max_target=100.0,
    # NOTE: setting noise=0 makes some comparisons with supervision=1.0 fail.
    # I suppose it can be due to two feature columns occasionally having the
    # same order, especially when there are few samples in the node.
    noise=0.1,  
)


# FIXME: very small values sometimes cause trees to differ, probably due to
# precision issues.
@pytest.fixture(params=[0.0, 0.0001, 0.1, 0.2328, 0.569, 0.782, 0.995, 1.0])
def supervision(request):
    return request.param


@pytest.fixture(params=range(3))
def random_state(request):
    return request.param


# @pytest.fixture(params=[None, 5])
@pytest.fixture(params=[1, 2, 3])
def max_depth(request):
    return request.param


# NOTE: small leaves may lead to differing trees by chance. Simply becase,
# with just a few samples, one feature column may ocasionally have values with
# the same order as another column, so that they yield the same split position
# and impurities. Since we cannot dictate the order in which features are
# chosen to be evaluated by the splitters, the two equally-ordered features
# can be swapped.
# @pytest.fixture(params=[1, 6, 10, .1])
@pytest.fixture(params=[0.05])
def msl(request):  # min_samples_leaf parameter
    return request.param


def test_monopartite_semisupervised(supervision, msl, random_state, **params):
    params = DEF_PARAMS | params

    tree_ss = DecisionTreeRegressorSS(
        supervision=supervision,
        min_samples_leaf=msl,
        random_state=random_state,
    )
    tree1 = DecisionTreeRegressor(
        min_samples_leaf=msl,
        random_state=random_state,
    )

    return compare_trees(
        tree1=tree1,
        tree2=tree_ss,
        tree2_is_2d=False,
        supervision=supervision,
        random_state=random_state,
        **params,
    )


def test_semisupervision_1d2d(
    supervision,
    random_state,
    msl,
    max_depth,
    **params,
):
    params = DEF_PARAMS | params
    # n_samples = np.prod(params['n_samples'])
    # max_depth = int(np.ceil(np.log2(n_samples/6))) if msl == 1 else None

    print('* Supervision level:', supervision)
    print('* Max depth:', max_depth)

    tree1 = DecisionTreeRegressorSS(
        supervision=supervision,
        min_samples_leaf=msl,
        random_state=random_state,
        max_depth=max_depth,
    )
    tree2 = BipartiteDecisionTreeRegressorSS(
        supervision=supervision,
        min_samples_leaf=msl,
        random_state=random_state,
        max_depth=max_depth,
    )

    return compare_trees(
        tree1=tree1,
        tree2=tree2,
        tree2_is_2d=True,
        supervision=1.0,  # tree1 will already apply the supervision
        random_state=random_state,
        **params,
    )


@pytest.mark.parametrize('update_supervision', [
    lambda **_: 0.7,
    lambda original_supervision, **_: original_supervision,
    lambda weighted_n_node_samples, weighted_n_samples, **_:
        1 - 0.3 * weighted_n_node_samples/weighted_n_samples,
])
def test_dynamic_supervision_1d2d(
    supervision, random_state, msl, max_depth, update_supervision, **params,
):
    params = DEF_PARAMS | params
    mono_sup_values = []
    bi_sup_values = []

    def mono_update_supervision(**kwargs):
        mono_sup_values.append(kwargs['current_supervision'])
        return update_supervision(**kwargs)
    def bi_update_supervision(**kwargs):
        bi_sup_values.append(kwargs['current_supervision'])
        return update_supervision(**kwargs)

    tree1 = DecisionTreeRegressorSS(
        criterion="squared_error",
        unsupervised_criterion="squared_error",
        supervision=supervision,
        update_supervision=mono_update_supervision,
        min_samples_leaf=msl,
        max_depth=max_depth,
    )
    tree2 = BipartiteDecisionTreeRegressorSS(
        criterion="squared_error",
        unsupervised_criterion_rows="squared_error",
        unsupervised_criterion_cols="squared_error",
        supervision=supervision,
        update_supervision=bi_update_supervision,
        min_samples_leaf=msl,
        max_depth=max_depth,
    )

    return compare_trees(
        tree1=tree1,
        tree2=tree2,
        tree2_is_2d=True,
        supervision=1.0,
        random_state=random_state,
        **params,
    )



from bipartite_learn.model_selection._validation import (
    multipartite_cross_validate,
)
from bipartite_learn.model_selection._split import (
    check_multipartite_cv,
    MultipartiteCrossValidator,
    make_multipartite_train_test_splitter,
    make_multipartite_kfold,
    multipartite_train_test_split,
)
from bipartite_learn.model_selection._search import MultipartiteGridSearchCV
from bipartite_learn.model_selection._search import MultipartiteRandomizedSearchCV


__all__ = [
    "split_train_test_nd",
    "make_multipartite_train_test_splitter",
    "multipartite_cross_validate",
    "MultipartiteCrossValidator",
    "MultipartiteGridSearchCV",
    "MultipartiteRandomizedSearchCV",
]



"""Experimental interfaces for datasets.
"""
from abc import ABCMeta, abstractmethod, abstractproperty
import numpy as np


class BaseData(metaclass=ABCMeta):
    @abstractmethod
    def check(self):
        """Check validity of the data provided."""


class BaseDataset(BaseData, metaclass=ABCMeta):
    def __init__(self, X, y=None, check=False):
        self.X = X
        self.y = y
        self.Xy = X, y
        if check:
            self.check()

    @abstractmethod
    def __getitem__(*indices):
        ...
    
    def check(self, estimator=None):
        """Check validity of the data provided."""
        return self

    @abstractproperty
    def n_samples(self):
        """The number of samples in the dataset."""

    @abstractproperty
    def n_features(self):
        """The number of features in the dataset."""


class StandardDataset(BaseDataset):
    def __getitem__(self, indices):
        return self.__class__(
            X=self.X[indices],
            y=self.y and self.y[indices],
        )

    @property
    def n_samples(self):
        return self.X.shape[0]

    @property
    def n_features(self):
        return self.X.shape[1]

    @property
    def n_outputs(self):
        return self.y.shape[1]


class PairwiseDataset(BaseDataset):
    def __getitem__(self, indices):
        return self.__class__(
            X=self.X[np.ix_(indices, indices)],
            y=self.y and self.y[indices],
        )


class BipartiteDataset(BaseDataset):
    def __getitem__(self, indices):
        X_ = [Xi[np.r_[idx]] for Xi, idx in zip(self.X, indices)]
        if self.y is None:
            y_ = None
        else:
            y_ = self.y[np.ix_(*(np.r_[idx] for idx in indices))]
        return self.__class__(X_, y_)

    @property
    def n_samples(self):
        # TODO: Use sklearn's _n_samples()
        return tuple(Xi.shape[0] for Xi in self.X)

    @property
    def n_features(self):
        return tuple(Xi.shape[1] for Xi in self.X)

    @property
    def n_parts(self):
        return len(self.X)


import logging
import numpy as np
import pytest
from sklearn.utils._testing import assert_allclose
from sklearn.ensemble._gb import (
    GradientBoostingClassifier,
    GradientBoostingRegressor,
)
from bipartite_learn.ensemble._gb import (
    BipartiteGradientBoostingClassifier,
    BipartiteGradientBoostingRegressor,
)
from bipartite_learn.wrappers import GlobalSingleOutputWrapper
from .utils.make_examples import make_interaction_regression
from .utils.test_utils import stopwatch


@pytest.fixture(params=range(5))
def random_state(request):
    return request.param


@pytest.mark.parametrize("subsample", (1.0, 0.1))
@pytest.mark.parametrize(
    "mono_criterion, bi_criterion",
    [
        ("squared_error", "squared_error_gso"),
        ("friedman_mse", "friedman_gso"),
    ],
    ids=['mse', 'friedman'],
)
def test_gradient_boosting(
    random_state,
    subsample,
    mono_criterion,
    bi_criterion,
):
    estimator1 = GradientBoostingRegressor(
        criterion=mono_criterion,
        random_state=random_state,
        subsample=subsample,
    )
    estimator = BipartiteGradientBoostingRegressor(
        criterion=bi_criterion,
        random_state=random_state,
        subsample=subsample,
        bipartite_adapter="gmosa",
    )

    XX, Y, X, y = make_interaction_regression(
        n_samples=(40, 50),
        random_state=random_state,
        return_molten=True,
    )
    y = y.reshape(-1, 1)

    logging.info(f"{random_state=}, {subsample=}")
    with stopwatch("Fitting bipartite..."):
        estimator.fit(XX, Y)
    with stopwatch("Fitting monopartite..."):
        estimator1.fit(X, y)

    pred1 = estimator1.predict(X)
    pred2 = estimator.predict(XX)
    mse1 = np.mean((y - pred1)**2)
    mse2 = np.mean((y - pred2)**2)

    logging.info(f"Base MSE (var):  {y.var()}")
    logging.info(f"Monopartite MSE: {mse1}")
    logging.info(f"Bipartite MSE:   {mse2}")

    if subsample == 1.0:
        assert_allclose(mse1, mse2)
        assert_allclose(pred1, pred2)



"""
This module gathers tree-based methods, including decision, regression and
randomized trees, adapted from sklearn for semi-supervised learning.
"""

# Author: Pedro Ilidio <pedrilidio@gmail.com>
# Adapted from scikit-learn.
#
# License: BSD 3 clause

from typing import Type
from copy import deepcopy
from abc import ABCMeta
from abc import abstractmethod
from math import ceil
from numbers import Integral, Real

import numpy as np
from scipy.sparse import issparse

from sklearn.base import is_classifier, _fit_context
from sklearn.utils import check_random_state
from sklearn.utils.validation import _check_sample_weight
from sklearn.utils import compute_sample_weight
from sklearn.utils.multiclass import check_classification_targets
from sklearn.utils._param_validation import (
    validate_params, Interval, StrOptions, Hidden,
)
from sklearn.metrics.pairwise import check_pairwise_arrays

from sklearn.tree._criterion import Criterion
from sklearn.tree._splitter import Splitter
from sklearn.tree._tree import (
    Tree, DepthFirstTreeBuilder, BestFirstTreeBuilder
)
from sklearn.tree import _tree
# Hypertree-specific:
from ._bipartite_splitter import BipartiteSplitter

# Semi-supervision-specific:
from sklearn.tree._classes import (
    BaseDecisionTree,
    DecisionTreeRegressor,
    DecisionTreeClassifier,
    DENSE_SPLITTERS,
    SPARSE_SPLITTERS,
)
from ._bipartite_classes import (
    BaseBipartiteDecisionTree,
    BipartiteDecisionTreeRegressor,
    BipartiteExtraTreeRegressor,
    _get_criterion_classes,
)
from . import (
    _semisupervised_criterion,
    _unsupervised_criterion,
    _splitter_factory,
    _axis_criterion,
)


__all__ = [
    "DecisionTreeClassifierSS",
    "ExtraTreeClassifierSS",
    "DecisionTreeRegressorSS",
    "ExtraTreeRegressorSS",
    "BipartiteDecisionTreeRegressorSS",
]


# =============================================================================
# Types and constants
# =============================================================================

DTYPE = _tree.DTYPE
DOUBLE = _tree.DOUBLE

# TODO: unsupervised splitters that consider the current feature column only.
# SINGLE_FEATURE_SPLITTERS = {
#     "best": _semisupervised_splitter.BestSplitterSFSS,
#     "random": _semisupervised_splitter.RandomSplitterSFSS,
# }

SS_CRITERIA = {
    "default": _semisupervised_criterion.SSCompositeCriterion,
}

# Unuspervised criteria for numeric X (based on regression criteria).
U_NUMERIC_CRITERIA = {
    "squared_error": _unsupervised_criterion.UnsupervisedSquaredError,
    "friedman": _unsupervised_criterion.UnsupervisedFriedman,
    "pairwise_squared_error": _unsupervised_criterion.PairwiseSquaredError,
    "pairwise_friedman": _unsupervised_criterion.PairwiseFriedman,
    "mean_distance": _unsupervised_criterion.MeanDistance
}

# Unuspervised criteria for categoric X (based on classification criteria).
U_CATEGORIC_CRITERIA = {
    "gini": _unsupervised_criterion.UnsupervisedGini,
    "entropy": _unsupervised_criterion.UnsupervisedEntropy,
    "pairwise_gini": _unsupervised_criterion.PairwiseGini,
    "pairwise_entropy": _unsupervised_criterion.PairwiseEntropy,
}
UNSUPERVISED_CRITERIA = U_CATEGORIC_CRITERIA | U_NUMERIC_CRITERIA

PAIRWISE_CRITERION_OPTIONS = {
    "mean_distance",
    "pairwise_gini",
    "pairwise_entropy",
    "pairwise_squared_error",
    "pairwise_friedman",
}


def _encode_classes(y, class_weight=None):
    check_classification_targets(y)
    classes = []
    n_classes = []
    y_encoded = np.zeros(y.shape, dtype=int)

    for k in range(y.shape[1]):
        classes_k, y_encoded[:, k] = np.unique(y[:, k], return_inverse=True)
        classes.append(classes_k)
        n_classes.append(classes_k.shape[0])

    if class_weight is None:
        expanded_class_weight = None
    else:
        expanded_class_weight = compute_sample_weight(class_weight, y)

    n_classes = np.array(n_classes, dtype=np.intp)

    return y_encoded, classes, n_classes, expanded_class_weight


@validate_params(
    {
        "criterion": [str, Criterion, type(Criterion)]
    },
    prefer_skip_nested_validation=True,
)
def _is_categoric_criterion(criterion: str | Criterion | Type[Criterion]):
    # TODO: different base classes for categoric (classification) and
    # regression (numeric) unsupervised criteria.
    if isinstance(criterion, str):
        return criterion in U_CATEGORIC_CRITERIA
    if isinstance(criterion, type):
        return issubclass(criterion, tuple(U_CATEGORIC_CRITERIA.values())),
    if isinstance(criterion, Criterion):
        return issubclass(criterion, tuple(U_CATEGORIC_CRITERIA.values())),
    raise TypeError


def _is_pairwise_criterion(unsupervised_criterion):
    return (
        unsupervised_criterion in PAIRWISE_CRITERION_OPTIONS
        or isinstance(
            unsupervised_criterion,
            _unsupervised_criterion.PairwiseCriterion
        )
    )


def _validate_X_targets(
    *,
    X,
    X_targets,
    is_categoric,
    is_pairwise,
):
    if X_targets is None:
        # _X_targets will be the target matrix for unsupervised criteria,
        # and thus must be formatted like y usually is, double-valued and
        # C contiguous.
        X_targets = X.toarray() if issparse(X) else X

        if is_categoric:
            # TODO: class weights for X.
            X_targets, *_ = _encode_classes(X_targets)

        if (
            getattr(X_targets, "dtype", None) != DOUBLE
            or not X_targets.flags.contiguous
        ):
            X_targets = np.ascontiguousarray(X_targets, dtype=DOUBLE)

    elif (
        getattr(X_targets, "dtype", None) != DOUBLE
        or not X_targets.flags.contiguous
    ):
        raise TypeError(
            "If provided, _X_targets must be a C contiguous array of float64"
        )

    if is_pairwise:
        check_pairwise_arrays(X_targets, Y=None, precomputed=True)

    return X_targets


def _validate_bipartite_X_targets(
    *,
    X,
    X_targets,
    is_categoric_rows,
    is_categoric_cols,
    is_pairwise_rows,
    is_pairwise_cols,
):
    if X_targets is None:
        X_targets = [None, None]

    for axis, (is_categoric, is_pairwise) in enumerate((
        (is_categoric_rows, is_pairwise_rows),
        (is_categoric_cols, is_pairwise_cols),
    )):
        X_targets[axis] = _validate_X_targets(
            X=X[axis],
            X_targets=X_targets[axis],
            is_categoric=is_categoric,
            is_pairwise=is_pairwise,
        )

    return X_targets


# =============================================================================
# Semisupervised classes
# =============================================================================


class BaseDecisionTreeSS(BaseDecisionTree, metaclass=ABCMeta):
    """Base class for decision trees.

    Warning: This class should not be used directly.
    Use derived classes instead.
    """
    _parameter_constraints: dict = {
        **BaseDecisionTree._parameter_constraints,
        "unsupervised_criterion": [
            StrOptions(set(UNSUPERVISED_CRITERIA.keys())),
            Hidden(Criterion),
        ],
        "supervision": [Interval(Real, 0.0, 1.0, closed="both")],
        "update_supervision": [callable, None],
        "ss_adapter": [
            None,
            Hidden(StrOptions(set(SS_CRITERIA.keys()))),
            Hidden(_semisupervised_criterion.SSCompositeCriterion),
        ],
        "preprocess_X_targets": [callable, None],
    }

    def _more_tags(self):
        # For cross-validation routines to split data correctly
        return {
            "pairwise": _is_pairwise_criterion(self.unsupervised_criterion)
        }

    @abstractmethod
    def __init__(
        self,
        *,
        supervision=0.5,
        unsupervised_criterion="squared_error",
        update_supervision=None,
        ss_adapter=None,
        preprocess_X_targets=None,
        _X_targets=None,
        **kwargs,
    ):
        self.supervision = supervision
        self.ss_adapter = ss_adapter
        self.unsupervised_criterion = unsupervised_criterion
        self.update_supervision = update_supervision
        self.preprocess_X_targets = preprocess_X_targets
        self._X_targets = _X_targets

        super().__init__(**kwargs)

    def _check_X_targets(self, X, _X_targets=None):
        if _X_targets is None and self.preprocess_X_targets is not None:
            X = self.preprocess_X_targets(X)

        return _validate_X_targets(
            X=X,
            X_targets=_X_targets,
            is_categoric=_is_categoric_criterion(
                self.unsupervised_criterion,
            ),
            is_pairwise=_is_pairwise_criterion(
                self.unsupervised_criterion,
            ),
        )

    # FIXME: Avoid copying the whole BaseDecisionTree.fit(). We currently do it
    # just so we can build the Splitter object our own way.
    @_fit_context(prefer_skip_nested_validation=True)
    def fit(self, X, y, sample_weight=None, check_input=True, _X_targets=None):
        random_state = check_random_state(self.random_state)

        if check_input:
            # TODO: Since we cannot pass _X_targets to trees in an ensemble,
            # inside sklearn.ensemble._forest._parallel_build_trees(),
            # we add _X_targets as a parameter in self.__init__ and in the
            # forest.estimator_params list, that will be reused by the
            # individual trees without copying the array for each tree
            # and consuming a large amount of memory.  In forests, check_inpu
            # is set to false, in which case we do not override self._X_targets
            # below.
            self._X_targets = self._check_X_targets(X, _X_targets)

            # Need to validate separately here.
            # We can't pass multi_output=True because that would allow y to be
            # csr.
            check_X_params = dict(dtype=DTYPE, accept_sparse="csc")
            check_y_params = dict(ensure_2d=False, dtype=None)
            X, y = self._validate_data(
                X, y, validate_separately=(check_X_params, check_y_params)
            )

            if self.criterion == "poisson":
                if np.any(y < 0):
                    raise ValueError(
                        "Some value(s) of y are negative which is"
                        " not allowed for Poisson regression."
                    )
                if np.sum(y) <= 0:
                    raise ValueError(
                        "Sum of y is not positive which is "
                        "necessary for Poisson regression."
                    )

        # Determine output settings
        n_samples, self.n_features_in_ = X.shape

        if len(y) != n_samples:
            raise ValueError(
                "Number of labels=%d does not match number of samples=%d"
                % (len(y), n_samples)
            )

        is_classification = is_classifier(self)

        y = np.atleast_1d(y)
        expanded_class_weight = None

        if y.ndim == 1:
            # reshape is necessary to preserve the data contiguity against vs
            # [:, np.newaxis] that does not.
            y = np.reshape(y, (-1, 1))

        self.n_outputs_ = y.shape[1]

        if is_classification:
            y, self.classes_, self.n_classes_, expanded_class_weight = \
                _encode_classes(y, self.class_weight)

        max_depth = np.iinfo(
            np.int32).max if self.max_depth is None else self.max_depth

        if isinstance(self.min_samples_leaf, Integral):
            min_samples_leaf = self.min_samples_leaf
        else:  # float
            min_samples_leaf = int(ceil(self.min_samples_leaf * n_samples))

        if isinstance(self.min_samples_split, Integral):
            min_samples_split = self.min_samples_split
        else:  # float
            min_samples_split = int(ceil(self.min_samples_split * n_samples))
            min_samples_split = max(2, min_samples_split)

        min_samples_split = max(min_samples_split, 2 * min_samples_leaf)

        if isinstance(self.max_features, str):
            if self.max_features == "sqrt":
                max_features = max(1, int(np.sqrt(self.n_features_in_)))
            elif self.max_features == "log2":
                max_features = max(1, int(np.log2(self.n_features_in_)))
        elif self.max_features is None:
            max_features = self.n_features_in_
        elif isinstance(self.max_features, Integral):
            max_features = self.max_features
        else:  # float
            if self.max_features > 0.0:
                max_features = max(
                    1, int(self.max_features * self.n_features_in_))
            else:
                max_features = 0

        self.max_features_ = max_features

        max_leaf_nodes = -1 if self.max_leaf_nodes is None else self.max_leaf_nodes

        if sample_weight is not None:
            sample_weight = _check_sample_weight(sample_weight, X, DOUBLE)

        if expanded_class_weight is not None:
            if sample_weight is not None:
                sample_weight = sample_weight * expanded_class_weight
            else:
                sample_weight = expanded_class_weight

        # Set min_weight_leaf from min_weight_fraction_leaf
        if sample_weight is None:
            min_weight_leaf = self.min_weight_fraction_leaf * n_samples
        else:
            min_weight_leaf = self.min_weight_fraction_leaf * \
                np.sum(sample_weight)

        # Build tree
        splitter = self._make_splitter(
            X=X,
            X_targets=self._X_targets,
            min_samples_leaf=min_samples_leaf,
            min_weight_leaf=min_weight_leaf,
            random_state=random_state,
        )

        if is_classifier(self):
            self.tree_ = Tree(
                self.n_features_in_,
                self.n_classes_,
                self.n_outputs_,
            )
        else:
            self.tree_ = Tree(
                self.n_features_in_,
                # TODO: tree shouldn't need this in this case
                np.array([1] * self.n_outputs_, dtype=np.intp),
                self.n_outputs_,
            )

        # Use BestFirst if max_leaf_nodes given; use DepthFirst otherwise
        if max_leaf_nodes < 0:
            builder = DepthFirstTreeBuilder(
                splitter,
                min_samples_split,
                min_samples_leaf,
                min_weight_leaf,
                max_depth,
                self.min_impurity_decrease,
            )
        else:
            builder = BestFirstTreeBuilder(
                splitter,
                min_samples_split,
                min_samples_leaf,
                min_weight_leaf,
                max_depth,
                max_leaf_nodes,
                self.min_impurity_decrease,
            )

        builder.build(self.tree_, X, y, sample_weight)

        if self.n_outputs_ == 1 and is_classifier(self):
            self.n_classes_ = self.n_classes_[0]
            self.classes_ = self.classes_[0]

        self._prune_tree()

        return self

    def _check_criterion(
        self,
        *,
        n_samples,
        X_targets=None,
        n_outputs=None,
        n_features=None,
        supervision=None,
        ss_adapter=None,
        criterion=None,
        unsupervised_criterion=None,
        update_supervision=None,
    ):
        # FIXME: classification is not covered.
        n_outputs = n_outputs or self.n_outputs_
        n_features = n_features or self.n_features_in_
        supervision = supervision or self.supervision
        ss_adapter = ss_adapter or self.ss_adapter
        criterion = criterion or self.criterion
        update_supervision = update_supervision or self.update_supervision
        X_targets = X_targets if X_targets is not None else self._X_targets
        unsupervised_criterion = unsupervised_criterion or self.unsupervised_criterion

        if isinstance(ss_adapter, str):
            ss_adapter = SS_CRITERIA[ss_adapter]
        elif ss_adapter is not None:  # SSCompositeCriterion:
            # Make a deepcopy in case the splitter has mutable attributes that
            # might be shared and modified concurrently during parallel fitting
            return deepcopy(ss_adapter)

        if isinstance(criterion, str):
            if is_classifier(self):
                CRITERIA = U_CATEGORIC_CRITERIA
            else:
                CRITERIA = U_NUMERIC_CRITERIA
            criterion = CRITERIA[criterion]
        else:
            criterion = deepcopy(criterion)

        if isinstance(unsupervised_criterion, str):
            unsupervised_criterion = UNSUPERVISED_CRITERIA[unsupervised_criterion]
        else:
            unsupervised_criterion = deepcopy(unsupervised_criterion)

        final_criterion = _splitter_factory.make_semisupervised_criterion(
            ss_class=ss_adapter,
            supervision=supervision,
            supervised_criterion=criterion,
            unsupervised_criterion=unsupervised_criterion,
            n_outputs=n_outputs,
            n_features=n_features,
            n_samples=n_samples,
            update_supervision=update_supervision,
        )

        if X_targets is None:
            raise RuntimeError("_X_targets was not set.")

        final_criterion.set_X(X_targets)
        return final_criterion

    def _make_splitter(
        self,
        *,
        X,
        X_targets,
        min_samples_leaf,
        min_weight_leaf,
        random_state,
    ):
        SPLITTERS = SPARSE_SPLITTERS if issparse(X) else DENSE_SPLITTERS

        if isinstance(self.splitter, Splitter):
            return deepcopy(self.splitter)
        else:
            criterion = self._check_criterion(
                X_targets=X_targets, n_samples=X.shape[0],
            )

            return SPLITTERS[self.splitter](
                criterion,
                self.max_features_,
                min_samples_leaf,
                min_weight_leaf,
                random_state,
            )


# =============================================================================
# Public estimators
# =============================================================================


class DecisionTreeClassifierSS(BaseDecisionTreeSS, DecisionTreeClassifier):
    """A decision tree classifier (semi-supervised version).

    Read more in the :ref:`User Guide <tree>`.

    Parameters
    ----------
    criterion : {"gini", "entropy", "log_loss"}, default="gini"
        The function to measure the quality of a split. Supported criteria are
        "gini" for the Gini impurity and "log_loss" and "entropy" both for the
        Shannon information gain, see :ref:`tree_mathematical_formulation`.

    splitter : {"best", "random"}, default="best"
        The strategy used to choose the split at each node. Supported
        strategies are "best" to choose the best split and "random" to choose
        the best random split.

    max_depth : int, default=None
        The maximum depth of the tree. If None, then nodes are expanded until
        all leaves are pure or until all leaves contain less than
        min_samples_split samples.

    min_samples_split : int or float, default=2
        The minimum number of samples required to split an internal node:

        - If int, then consider `min_samples_split` as the minimum number.
        - If float, then `min_samples_split` is a fraction and
          `ceil(min_samples_split * n_samples)` are the minimum
          number of samples for each split.

        .. versionchanged:: 0.18
           Added float values for fractions.

    min_samples_leaf : int or float, default=1
        The minimum number of samples required to be at a leaf node.
        A split point at any depth will only be considered if it leaves at
        least ``min_samples_leaf`` training samples in each of the left and
        right branches.  This may have the effect of smoothing the model,
        especially in regression.

        - If int, then consider `min_samples_leaf` as the minimum number.
        - If float, then `min_samples_leaf` is a fraction and
          `ceil(min_samples_leaf * n_samples)` are the minimum
          number of samples for each node.

        .. versionchanged:: 0.18
           Added float values for fractions.

    min_weight_fraction_leaf : float, default=0.0
        The minimum weighted fraction of the sum total of weights (of all
        the input samples) required to be at a leaf node. Samples have
        equal weight when sample_weight is not provided.

    max_features : int, float or {"auto", "sqrt", "log2"}, default=None
        The number of features to consider when looking for the best split:

            - If int, then consider `max_features` features at each split.
            - If float, then `max_features` is a fraction and
              `int(max_features * n_features)` features are considered at each
              split.
            - If "auto", then `max_features=sqrt(n_features)`.
            - If "sqrt", then `max_features=sqrt(n_features)`.
            - If "log2", then `max_features=log2(n_features)`.
            - If None or 1.0, then `max_features=n_features`.

            .. deprecated:: 1.1
                The `"auto"` option was deprecated in 1.1 and will be removed
                in 1.3.

        Note: the search for a split does not stop until at least one
        valid partition of the node samples is found, even if it requires to
        effectively inspect more than ``max_features`` features.

    random_state : int, RandomState instance or None, default=None
        Controls the randomness of the estimator. The features are always
        randomly permuted at each split, even if ``splitter`` is set to
        ``"best"``. When ``max_features < n_features``, the algorithm will
        select ``max_features`` at random at each split before finding the best
        split among them. But the best found split may vary across different
        runs, even if ``max_features=n_features``. That is the case, if the
        improvement of the criterion is identical for several splits and one
        split has to be selected at random. To obtain a deterministic behaviour
        during fitting, ``random_state`` has to be fixed to an integer.
        See :term:`Glossary <random_state>` for details.

    max_leaf_nodes : int, default=None
        Grow a tree with ``max_leaf_nodes`` in best-first fashion.
        Best nodes are defined as relative reduction in impurity.
        If None then unlimited number of leaf nodes.

    min_impurity_decrease : float, default=0.0
        A node will be split if this split induces a decrease of the impurity
        greater than or equal to this value.

        The weighted impurity decrease equation is the following::

            N_t / N * (impurity - N_t_R / N_t * right_impurity
                                - N_t_L / N_t * left_impurity)

        where ``N`` is the total number of samples, ``N_t`` is the number of
        samples at the current node, ``N_t_L`` is the number of samples in the
        left child, and ``N_t_R`` is the number of samples in the right child.

        ``N``, ``N_t``, ``N_t_R`` and ``N_t_L`` all refer to the weighted sum,
        if ``sample_weight`` is passed.

        .. versionadded:: 0.19

    class_weight : dict, list of dict or "balanced", default=None
        Weights associated with classes in the form ``{class_label: weight}``.
        If None, all classes are supposed to have weight one. For
        multi-output problems, a list of dicts can be provided in the same
        order as the columns of y.

        Note that for multioutput (including multilabel) weights should be
        defined for each class of every column in its own dict. For example,
        for four-class multilabel classification weights should be
        [{0: 1, 1: 1}, {0: 1, 1: 5}, {0: 1, 1: 1}, {0: 1, 1: 1}] instead of
        [{1:1}, {2:5}, {3:1}, {4:1}].

        The "balanced" mode uses the values of y to automatically adjust
        weights inversely proportional to class frequencies in the input data
        as ``n_samples / (n_classes * np.bincount(y))``

        For multi-output, the weights of each column of y will be multiplied.

        Note that these weights will be multiplied with sample_weight (passed
        through the fit method) if sample_weight is specified.

    ccp_alpha : non-negative float, default=0.0
        Complexity parameter used for Minimal Cost-Complexity Pruning. The
        subtree with the largest cost complexity that is smaller than
        ``ccp_alpha`` will be chosen. By default, no pruning is performed. See
        :ref:`minimal_cost_complexity_pruning` for details.

        .. versionadded:: 0.22

    Attributes
    ----------
    classes_ : ndarray of shape (n_classes,) or list of ndarray
        The classes labels (single output problem),
        or a list of arrays of class labels (multi-output problem).

    feature_importances_ : ndarray of shape (n_features,)
        The impurity-based feature importances.
        The higher, the more important the feature.
        The importance of a feature is computed as the (normalized)
        total reduction of the criterion brought by that feature.  It is also
        known as the Gini importance [4]_.

        Warning: impurity-based feature importances can be misleading for
        high cardinality features (many unique values). See
        :func:`sklearn.inspection.permutation_importance` as an alternative.

    max_features_ : int
        The inferred value of max_features.

    n_classes_ : int or list of int
        The number of classes (for single output problems),
        or a list containing the number of classes for each
        output (for multi-output problems).

    n_features_in_ : int
        Number of features seen during :term:`fit`.

        .. versionadded:: 0.24

    feature_names_in_ : ndarray of shape (`n_features_in_`,)
        Names of features seen during :term:`fit`. Defined only when `X`
        has feature names that are all strings.

        .. versionadded:: 1.0

    n_outputs_ : int
        The number of outputs when ``fit`` is performed.

    tree_ : Tree instance
        The underlying Tree object. Please refer to
        ``help(sklearn.tree._tree.Tree)`` for attributes of Tree object and
        :ref:`sphx_glr_auto_examples_tree_plot_unveil_tree_structure.py`
        for basic usage of these attributes.

    See Also
    --------
    DecisionTreeRegressor : A decision tree regressor.

    Notes
    -----
    The default values for the parameters controlling the size of the trees
    (e.g. ``max_depth``, ``min_samples_leaf``, etc.) lead to fully grown and
    unpruned trees which can potentially be very large on some data sets. To
    reduce memory consumption, the complexity and size of the trees should be
    controlled by setting those parameter values.

    The :meth:`predict` method operates using the :func:`numpy.argmax`
    function on the outputs of :meth:`predict_proba`. This means that in
    case the highest predicted probabilities are tied, the classifier will
    predict the tied class with the lowest index in :term:`classes_`.

    References
    ----------

    .. [1] https://en.wikipedia.org/wiki/Decision_tree_learning

    .. [2] L. Breiman, J. Friedman, R. Olshen, and C. Stone, "Classification
           and Regression Trees", Wadsworth, Belmont, CA, 1984.

    .. [3] T. Hastie, R. Tibshirani and J. Friedman. "Elements of Statistical
           Learning", Springer, 2009.

    .. [4] L. Breiman, and A. Cutler, "Random Forests",
           https://www.stat.berkeley.edu/~breiman/RandomForests/cc_home.htm

    Examples
    --------
    >>> from sklearn.datasets import load_iris
    >>> from sklearn.model_selection import cross_val_score
    >>> from sklearn.tree import DecisionTreeClassifier
    >>> clf = DecisionTreeClassifier(random_state=0)
    >>> iris = load_iris()
    >>> cross_val_score(clf, iris.data, iris.target, cv=10)
    ...                             # doctest: +SKIP
    ...
    array([ 1.     ,  0.93...,  0.86...,  0.93...,  0.93...,
            0.93...,  0.93...,  1.     ,  0.93...,  1.      ])
    """

    _parameter_constraints: dict = {
        **BaseDecisionTreeSS._parameter_constraints,
        **DecisionTreeClassifier._parameter_constraints,
    }
    _parameter_constraints["criterion"] = [
        StrOptions(
            set(U_CATEGORIC_CRITERIA.keys())
            - PAIRWISE_CRITERION_OPTIONS,
        ),
        Hidden(_axis_criterion.BaseComposableCriterion),
    ]
    _parameter_constraints.pop("splitter")

    def __init__(
        self,
        *,
        criterion="gini",
        splitter="best",
        max_depth=None,
        min_samples_split=2,
        min_samples_leaf=1,
        min_weight_fraction_leaf=0.0,
        max_features=None,
        random_state=None,
        max_leaf_nodes=None,
        min_impurity_decrease=0.0,
        class_weight=None,
        ccp_alpha=0.0,
        # Semi-supervised parameters:
        supervision=0.5,
        ss_adapter=None,
        unsupervised_criterion="squared_error",
        update_supervision=None,
        preprocess_X_targets=None,
        _X_targets=None,
    ):
        super().__init__(
            splitter=splitter,
            criterion=criterion,
            max_depth=max_depth,
            min_samples_split=min_samples_split,
            min_samples_leaf=min_samples_leaf,
            min_weight_fraction_leaf=min_weight_fraction_leaf,
            max_features=max_features,
            max_leaf_nodes=max_leaf_nodes,
            class_weight=class_weight,
            random_state=random_state,
            min_impurity_decrease=min_impurity_decrease,
            ccp_alpha=ccp_alpha,
            # Semi-supervised parameters:
            supervision=supervision,
            ss_adapter=ss_adapter,
            unsupervised_criterion=unsupervised_criterion,
            update_supervision=update_supervision,
            preprocess_X_targets=preprocess_X_targets,
            _X_targets=_X_targets,
        )


class DecisionTreeRegressorSS(BaseDecisionTreeSS, DecisionTreeRegressor):
    """A decision tree regressor (semi-supervised version).

    Read more in the :ref:`User Guide <tree>`.

    Parameters
    ----------
    criterion : {"squared_error", "friedman_mse", "absolute_error", \
            "poisson"}, default="squared_error"
        The function to measure the quality of a split. Supported criteria
        are "squared_error" for the mean squared error, which is equal to
        variance reduction as feature selection criterion and minimizes the L2
        loss using the mean of each terminal node, "friedman_mse", which uses
        mean squared error with Friedman's improvement score for potential
        splits, "absolute_error" for the mean absolute error, which minimizes
        the L1 loss using the median of each terminal node, and "poisson" which
        uses reduction in Poisson deviance to find splits.

        .. versionadded:: 0.18
           Mean Absolute Error (MAE) criterion.

        .. versionadded:: 0.24
            Poisson deviance criterion.

        .. deprecated:: 1.0
            Criterion "mse" was deprecated in v1.0 and will be removed in
            version 1.2. Use `criterion="squared_error"` which is equivalent.

        .. deprecated:: 1.0
            Criterion "mae" was deprecated in v1.0 and will be removed in
            version 1.2. Use `criterion="absolute_error"` which is equivalent.

    splitter : {"best", "random"}, default="best"
        The strategy used to choose the split at each node. Supported
        strategies are "best" to choose the best split and "random" to choose
        the best random split.

    max_depth : int, default=None
        The maximum depth of the tree. If None, then nodes are expanded until
        all leaves are pure or until all leaves contain less than
        min_samples_split samples.

    min_samples_split : int or float, default=2
        The minimum number of samples required to split an internal node:

        - If int, then consider `min_samples_split` as the minimum number.
        - If float, then `min_samples_split` is a fraction and
          `ceil(min_samples_split * n_samples)` are the minimum
          number of samples for each split.

        .. versionchanged:: 0.18
           Added float values for fractions.

    min_samples_leaf : int or float, default=1
        The minimum number of samples required to be at a leaf node.
        A split point at any depth will only be considered if it leaves at
        least ``min_samples_leaf`` training samples in each of the left and
        right branches.  This may have the effect of smoothing the model,
        especially in regression.

        - If int, then consider `min_samples_leaf` as the minimum number.
        - If float, then `min_samples_leaf` is a fraction and
          `ceil(min_samples_leaf * n_samples)` are the minimum
          number of samples for each node.

        .. versionchanged:: 0.18
           Added float values for fractions.

    min_weight_fraction_leaf : float, default=0.0
        The minimum weighted fraction of the sum total of weights (of all
        the input samples) required to be at a leaf node. Samples have
        equal weight when sample_weight is not provided.

    max_features : int, float or {"auto", "sqrt", "log2"}, default=None
        The number of features to consider when looking for the best split:

        - If int, then consider `max_features` features at each split.
        - If float, then `max_features` is a fraction and
          `int(max_features * n_features)` features are considered at each
          split.
        - If "auto", then `max_features=n_features`.
        - If "sqrt", then `max_features=sqrt(n_features)`.
        - If "log2", then `max_features=log2(n_features)`.
        - If None or 1.0, then `max_features=n_features`.

        .. deprecated:: 1.1
            The `"auto"` option was deprecated in 1.1 and will be removed
            in 1.3.

        Note: the search for a split does not stop until at least one
        valid partition of the node samples is found, even if it requires to
        effectively inspect more than ``max_features`` features.

    random_state : int, RandomState instance or None, default=None
        Controls the randomness of the estimator. The features are always
        randomly permuted at each split, even if ``splitter`` is set to
        ``"best"``. When ``max_features < n_features``, the algorithm will
        select ``max_features`` at random at each split before finding the best
        split among them. But the best found split may vary across different
        runs, even if ``max_features=n_features``. That is the case, if the
        improvement of the criterion is identical for several splits and one
        split has to be selected at random. To obtain a deterministic behaviour
        during fitting, ``random_state`` has to be fixed to an integer.
        See :term:`Glossary <random_state>` for details.

    max_leaf_nodes : int, default=None
        Grow a tree with ``max_leaf_nodes`` in best-first fashion.
        Best nodes are defined as relative reduction in impurity.
        If None then unlimited number of leaf nodes.

    min_impurity_decrease : float, default=0.0
        A node will be split if this split induces a decrease of the impurity
        greater than or equal to this value.

        The weighted impurity decrease equation is the following::

            N_t / N * (impurity - N_t_R / N_t * right_impurity
                                - N_t_L / N_t * left_impurity)

        where ``N`` is the total number of samples, ``N_t`` is the number of
        samples at the current node, ``N_t_L`` is the number of samples in the
        left child, and ``N_t_R`` is the number of samples in the right child.

        ``N``, ``N_t``, ``N_t_R`` and ``N_t_L`` all refer to the weighted sum,
        if ``sample_weight`` is passed.

        .. versionadded:: 0.19

    ccp_alpha : non-negative float, default=0.0
        Complexity parameter used for Minimal Cost-Complexity Pruning. The
        subtree with the largest cost complexity that is smaller than
        ``ccp_alpha`` will be chosen. By default, no pruning is performed. See
        :ref:`minimal_cost_complexity_pruning` for details.

        .. versionadded:: 0.22

    Attributes
    ----------
    feature_importances_ : ndarray of shape (n_features,)
        The feature importances.
        The higher, the more important the feature.
        The importance of a feature is computed as the
        (normalized) total reduction of the criterion brought
        by that feature. It is also known as the Gini importance [4]_.

        Warning: impurity-based feature importances can be misleading for
        high cardinality features (many unique values). See
        :func:`sklearn.inspection.permutation_importance` as an alternative.

    max_features_ : int
        The inferred value of max_features.

    n_features_in_ : int
        Number of features seen during :term:`fit`.

    feature_names_in_ : ndarray of shape (`n_features_in_`,)
        Names of features seen during :term:`fit`. Defined only when `X`
        has feature names that are all strings.

    n_outputs_ : int
        The number of outputs when ``fit`` is performed.

    tree_ : Tree instance
        The underlying Tree object. Please refer to
        ``help(sklearn.tree._tree.Tree)`` for attributes of Tree object and
        :ref:`sphx_glr_auto_examples_tree_plot_unveil_tree_structure.py`
        for basic usage of these attributes.

    See Also
    --------
    DecisionTreeClassifier : A decision tree classifier.

    Notes
    -----
    The default values for the parameters controlling the size of the trees
    (e.g. ``max_depth``, ``min_samples_leaf``, etc.) lead to fully grown and
    unpruned trees which can potentially be very large on some data sets. To
    reduce memory consumption, the complexity and size of the trees should be
    controlled by setting those parameter values.

    References
    ----------

    .. [1] https://en.wikipedia.org/wiki/Decision_tree_learning

    .. [2] L. Breiman, J. Friedman, R. Olshen, and C. Stone, "Classification
           and Regression Trees", Wadsworth, Belmont, CA, 1984.

    .. [3] T. Hastie, R. Tibshirani and J. Friedman. "Elements of Statistical
           Learning", Springer, 2009.

    .. [4] L. Breiman, and A. Cutler, "Random Forests",
           https://www.stat.berkeley.edu/~breiman/RandomForests/cc_home.htm

    Examples
    --------
    >>> from sklearn.datasets import load_diabetes
    >>> from sklearn.model_selection import cross_val_score
    >>> from sklearn.tree import DecisionTreeRegressor
    >>> X, y = load_diabetes(return_X_y=True)
    >>> regressor = DecisionTreeRegressor(random_state=0)
    >>> cross_val_score(regressor, X, y, cv=10)
    ...                    # doctest: +SKIP
    ...
    array([-0.39..., -0.46...,  0.02...,  0.06..., -0.50...,
           0.16...,  0.11..., -0.73..., -0.30..., -0.00...])
    """

    _parameter_constraints: dict = {
        **BaseDecisionTreeSS._parameter_constraints,
        **DecisionTreeRegressor._parameter_constraints,
    }
    _parameter_constraints["criterion"] = [
        StrOptions(
            set(U_NUMERIC_CRITERIA.keys())
            - PAIRWISE_CRITERION_OPTIONS,
        ),
        Hidden(_axis_criterion.BaseComposableCriterion),
    ]
    _parameter_constraints.pop("splitter")

    def __init__(
        self,
        *,
        criterion="squared_error",
        splitter="best",
        max_depth=None,
        min_samples_split=2,
        min_samples_leaf=1,
        min_weight_fraction_leaf=0.0,
        max_features=None,
        random_state=None,
        max_leaf_nodes=None,
        min_impurity_decrease=0.0,
        ccp_alpha=0.0,
        # Semi-supervised parameters:
        supervision=0.5,
        ss_adapter=None,
        unsupervised_criterion="squared_error",
        update_supervision=None,
        preprocess_X_targets=None,
        _X_targets=None,
    ):
        super().__init__(
            splitter=splitter,
            criterion=criterion,
            max_depth=max_depth,
            min_samples_split=min_samples_split,
            min_samples_leaf=min_samples_leaf,
            min_weight_fraction_leaf=min_weight_fraction_leaf,
            max_features=max_features,
            max_leaf_nodes=max_leaf_nodes,
            random_state=random_state,
            min_impurity_decrease=min_impurity_decrease,
            ccp_alpha=ccp_alpha,
            # Semi-supervised parameters:
            supervision=supervision,
            ss_adapter=ss_adapter,
            unsupervised_criterion=unsupervised_criterion,
            update_supervision=update_supervision,
            preprocess_X_targets=preprocess_X_targets,
            _X_targets=_X_targets,
        )


class ExtraTreeClassifierSS(DecisionTreeClassifierSS):
    """An extremely randomized tree classifier (semi-supervised version).

    Extra-trees differ from classic decision trees in the way they are built.
    When looking for the best split to separate the samples of a node into two
    groups, random splits are drawn for each of the `max_features` randomly
    selected features and the best split among those is chosen. When
    `max_features` is set 1, this amounts to building a totally random
    decision tree.

    Warning: Extra-trees should only be used within ensemble methods.

    Read more in the :ref:`User Guide <tree>`.

    Parameters
    ----------
    criterion : {"gini", "entropy", "log_loss"}, default="gini"
        The function to measure the quality of a split. Supported criteria are
        "gini" for the Gini impurity and "log_loss" and "entropy" both for the
        Shannon information gain, see :ref:`tree_mathematical_formulation`.

    splitter : {"random", "best"}, default="random"
        The strategy used to choose the split at each node. Supported
        strategies are "best" to choose the best split and "random" to choose
        the best random split.

    max_depth : int, default=None
        The maximum depth of the tree. If None, then nodes are expanded until
        all leaves are pure or until all leaves contain less than
        min_samples_split samples.

    min_samples_split : int or float, default=2
        The minimum number of samples required to split an internal node:

        - If int, then consider `min_samples_split` as the minimum number.
        - If float, then `min_samples_split` is a fraction and
          `ceil(min_samples_split * n_samples)` are the minimum
          number of samples for each split.

        .. versionchanged:: 0.18
           Added float values for fractions.

    min_samples_leaf : int or float, default=1
        The minimum number of samples required to be at a leaf node.
        A split point at any depth will only be considered if it leaves at
        least ``min_samples_leaf`` training samples in each of the left and
        right branches.  This may have the effect of smoothing the model,
        especially in regression.

        - If int, then consider `min_samples_leaf` as the minimum number.
        - If float, then `min_samples_leaf` is a fraction and
          `ceil(min_samples_leaf * n_samples)` are the minimum
          number of samples for each node.

        .. versionchanged:: 0.18
           Added float values for fractions.

    min_weight_fraction_leaf : float, default=0.0
        The minimum weighted fraction of the sum total of weights (of all
        the input samples) required to be at a leaf node. Samples have
        equal weight when sample_weight is not provided.

    max_features : int, float, {"auto", "sqrt", "log2"} or None, default="sqrt"
        The number of features to consider when looking for the best split:

            - If int, then consider `max_features` features at each split.
            - If float, then `max_features` is a fraction and
              `int(max_features * n_features)` features are considered at each
              split.
            - If "auto", then `max_features=sqrt(n_features)`.
            - If "sqrt", then `max_features=sqrt(n_features)`.
            - If "log2", then `max_features=log2(n_features)`.
            - If None, then `max_features=n_features`.

            .. versionchanged:: 1.1
                The default of `max_features` changed from `"auto"` to `"sqrt"`.

            .. deprecated:: 1.1
                The `"auto"` option was deprecated in 1.1 and will be removed
                in 1.3.

        Note: the search for a split does not stop until at least one
        valid partition of the node samples is found, even if it requires to
        effectively inspect more than ``max_features`` features.

    random_state : int, RandomState instance or None, default=None
        Used to pick randomly the `max_features` used at each split.
        See :term:`Glossary <random_state>` for details.

    max_leaf_nodes : int, default=None
        Grow a tree with ``max_leaf_nodes`` in best-first fashion.
        Best nodes are defined as relative reduction in impurity.
        If None then unlimited number of leaf nodes.

    min_impurity_decrease : float, default=0.0
        A node will be split if this split induces a decrease of the impurity
        greater than or equal to this value.

        The weighted impurity decrease equation is the following::

            N_t / N * (impurity - N_t_R / N_t * right_impurity
                                - N_t_L / N_t * left_impurity)

        where ``N`` is the total number of samples, ``N_t`` is the number of
        samples at the current node, ``N_t_L`` is the number of samples in the
        left child, and ``N_t_R`` is the number of samples in the right child.

        ``N``, ``N_t``, ``N_t_R`` and ``N_t_L`` all refer to the weighted sum,
        if ``sample_weight`` is passed.

        .. versionadded:: 0.19

    class_weight : dict, list of dict or "balanced", default=None
        Weights associated with classes in the form ``{class_label: weight}``.
        If None, all classes are supposed to have weight one. For
        multi-output problems, a list of dicts can be provided in the same
        order as the columns of y.

        Note that for multioutput (including multilabel) weights should be
        defined for each class of every column in its own dict. For example,
        for four-class multilabel classification weights should be
        [{0: 1, 1: 1}, {0: 1, 1: 5}, {0: 1, 1: 1}, {0: 1, 1: 1}] instead of
        [{1:1}, {2:5}, {3:1}, {4:1}].

        The "balanced" mode uses the values of y to automatically adjust
        weights inversely proportional to class frequencies in the input data
        as ``n_samples / (n_classes * np.bincount(y))``

        For multi-output, the weights of each column of y will be multiplied.

        Note that these weights will be multiplied with sample_weight (passed
        through the fit method) if sample_weight is specified.

    ccp_alpha : non-negative float, default=0.0
        Complexity parameter used for Minimal Cost-Complexity Pruning. The
        subtree with the largest cost complexity that is smaller than
        ``ccp_alpha`` will be chosen. By default, no pruning is performed. See
        :ref:`minimal_cost_complexity_pruning` for details.

        .. versionadded:: 0.22

    Attributes
    ----------
    classes_ : ndarray of shape (n_classes,) or list of ndarray
        The classes labels (single output problem),
        or a list of arrays of class labels (multi-output problem).

    max_features_ : int
        The inferred value of max_features.

    n_classes_ : int or list of int
        The number of classes (for single output problems),
        or a list containing the number of classes for each
        output (for multi-output problems).

    feature_importances_ : ndarray of shape (n_features,)
        The impurity-based feature importances.
        The higher, the more important the feature.
        The importance of a feature is computed as the (normalized)
        total reduction of the criterion brought by that feature.  It is also
        known as the Gini importance.

        Warning: impurity-based feature importances can be misleading for
        high cardinality features (many unique values). See
        :func:`sklearn.inspection.permutation_importance` as an alternative.

    n_features_in_ : int
        Number of features seen during :term:`fit`.

    feature_names_in_ : ndarray of shape (`n_features_in_`,)
        Names of features seen during :term:`fit`. Defined only when `X`
        has feature names that are all strings.

    n_outputs_ : int
        The number of outputs when ``fit`` is performed.

    tree_ : Tree instance
        The underlying Tree object. Please refer to
        ``help(sklearn.tree._tree.Tree)`` for attributes of Tree object and
        :ref:`sphx_glr_auto_examples_tree_plot_unveil_tree_structure.py`
        for basic usage of these attributes.

    See Also
    --------
    ExtraTreeRegressor : An extremely randomized tree regressor.
    sklearn.ensemble.ExtraTreesClassifier : An extra-trees classifier.
    sklearn.ensemble.ExtraTreesRegressor : An extra-trees regressor.
    sklearn.ensemble.RandomForestClassifier : A random forest classifier.
    sklearn.ensemble.RandomForestRegressor : A random forest regressor.
    sklearn.ensemble.RandomTreesEmbedding : An ensemble of
        totally random trees.

    Notes
    -----
    The default values for the parameters controlling the size of the trees
    (e.g. ``max_depth``, ``min_samples_leaf``, etc.) lead to fully grown and
    unpruned trees which can potentially be very large on some data sets. To
    reduce memory consumption, the complexity and size of the trees should be
    controlled by setting those parameter values.

    References
    ----------

    .. [1] P. Geurts, D. Ernst., and L. Wehenkel, "Extremely randomized trees",
           Machine Learning, 63(1), 3-42, 2006.

    Examples
    --------
    >>> from sklearn.datasets import load_iris
    >>> from sklearn.model_selection import train_test_split
    >>> from sklearn.ensemble import BaggingClassifier
    >>> from sklearn.tree import ExtraTreeClassifier
    >>> X, y = load_iris(return_X_y=True)
    >>> X_train, X_test, y_train, y_test = train_test_split(
    ...    X, y, random_state=0)
    >>> extra_tree = ExtraTreeClassifier(random_state=0)
    >>> cls = BaggingClassifier(extra_tree, random_state=0).fit(
    ...    X_train, y_train)
    >>> cls.score(X_test, y_test)
    0.8947...
    """

    def __init__(
        self,
        *,
        criterion="gini",
        splitter="random",
        max_depth=None,
        min_samples_split=2,
        min_samples_leaf=1,
        min_weight_fraction_leaf=0.0,
        max_features="sqrt",
        random_state=None,
        max_leaf_nodes=None,
        min_impurity_decrease=0.0,
        class_weight=None,
        ccp_alpha=0.0,
        # Semi-supervised parameters:
        supervision=0.5,
        ss_adapter=None,
        unsupervised_criterion="squared_error",
        update_supervision=None,
        preprocess_X_targets=None,
        _X_targets=None,
    ):
        super().__init__(
            splitter=splitter,
            max_depth=max_depth,
            min_samples_split=min_samples_split,
            min_samples_leaf=min_samples_leaf,
            min_weight_fraction_leaf=min_weight_fraction_leaf,
            max_features=max_features,
            max_leaf_nodes=max_leaf_nodes,
            class_weight=class_weight,
            min_impurity_decrease=min_impurity_decrease,
            random_state=random_state,
            ccp_alpha=ccp_alpha,
            # Semi-supervised parameters:
            supervision=supervision,
            ss_adapter=ss_adapter,
            criterion=criterion,
            unsupervised_criterion=unsupervised_criterion,
            update_supervision=update_supervision,
            preprocess_X_targets=preprocess_X_targets,
            _X_targets=_X_targets,
        )


class ExtraTreeRegressorSS(DecisionTreeRegressorSS):
    """An extremely randomized tree regressor (semi-supervised version).

    Extra-trees differ from classic decision trees in the way they are built.
    When looking for the best split to separate the samples of a node into two
    groups, random splits are drawn for each of the `max_features` randomly
    selected features and the best split among those is chosen. When
    `max_features` is set 1, this amounts to building a totally random
    decision tree.

    Warning: Extra-trees should only be used within ensemble methods.

    Read more in the :ref:`User Guide <tree>`.

    Parameters
    ----------
    criterion : {"squared_error", "friedman_mse"}, default="squared_error"
        The function to measure the quality of a split. Supported criteria
        are "squared_error" for the mean squared error, which is equal to
        variance reduction as feature selection criterion and "mae" for the
        mean absolute error.

        .. versionadded:: 0.18
           Mean Absolute Error (MAE) criterion.

        .. versionadded:: 0.24
            Poisson deviance criterion.

        .. deprecated:: 1.0
            Criterion "mse" was deprecated in v1.0 and will be removed in
            version 1.2. Use `criterion="squared_error"` which is equivalent.

        .. deprecated:: 1.0
            Criterion "mae" was deprecated in v1.0 and will be removed in
            version 1.2. Use `criterion="absolute_error"` which is equivalent.

    splitter : {"random", "best"}, default="random"
        The strategy used to choose the split at each node. Supported
        strategies are "best" to choose the best split and "random" to choose
        the best random split.

    max_depth : int, default=None
        The maximum depth of the tree. If None, then nodes are expanded until
        all leaves are pure or until all leaves contain less than
        min_samples_split samples.

    min_samples_split : int or float, default=2
        The minimum number of samples required to split an internal node:

        - If int, then consider `min_samples_split` as the minimum number.
        - If float, then `min_samples_split` is a fraction and
          `ceil(min_samples_split * n_samples)` are the minimum
          number of samples for each split.

        .. versionchanged:: 0.18
           Added float values for fractions.

    min_samples_leaf : int or float, default=1
        The minimum number of samples required to be at a leaf node.
        A split point at any depth will only be considered if it leaves at
        least ``min_samples_leaf`` training samples in each of the left and
        right branches.  This may have the effect of smoothing the model,
        especially in regression.

        - If int, then consider `min_samples_leaf` as the minimum number.
        - If float, then `min_samples_leaf` is a fraction and
          `ceil(min_samples_leaf * n_samples)` are the minimum
          number of samples for each node.

        .. versionchanged:: 0.18
           Added float values for fractions.

    min_weight_fraction_leaf : float, default=0.0
        The minimum weighted fraction of the sum total of weights (of all
        the input samples) required to be at a leaf node. Samples have
        equal weight when sample_weight is not provided.

    max_features : int, float, {"auto", "sqrt", "log2"} or None, default=1.0
        The number of features to consider when looking for the best split:

        - If int, then consider `max_features` features at each split.
        - If float, then `max_features` is a fraction and
          `int(max_features * n_features)` features are considered at each
          split.
        - If "auto", then `max_features=n_features`.
        - If "sqrt", then `max_features=sqrt(n_features)`.
        - If "log2", then `max_features=log2(n_features)`.
        - If None, then `max_features=n_features`.

        .. versionchanged:: 1.1
            The default of `max_features` changed from `"auto"` to `1.0`.

        .. deprecated:: 1.1
            The `"auto"` option was deprecated in 1.1 and will be removed
            in 1.3.

        Note: the search for a split does not stop until at least one
        valid partition of the node samples is found, even if it requires to
        effectively inspect more than ``max_features`` features.

    random_state : int, RandomState instance or None, default=None
        Used to pick randomly the `max_features` used at each split.
        See :term:`Glossary <random_state>` for details.

    min_impurity_decrease : float, default=0.0
        A node will be split if this split induces a decrease of the impurity
        greater than or equal to this value.

        The weighted impurity decrease equation is the following::

            N_t / N * (impurity - N_t_R / N_t * right_impurity
                                - N_t_L / N_t * left_impurity)

        where ``N`` is the total number of samples, ``N_t`` is the number of
        samples at the current node, ``N_t_L`` is the number of samples in the
        left child, and ``N_t_R`` is the number of samples in the right child.

        ``N``, ``N_t``, ``N_t_R`` and ``N_t_L`` all refer to the weighted sum,
        if ``sample_weight`` is passed.

        .. versionadded:: 0.19

    max_leaf_nodes : int, default=None
        Grow a tree with ``max_leaf_nodes`` in best-first fashion.
        Best nodes are defined as relative reduction in impurity.
        If None then unlimited number of leaf nodes.

    ccp_alpha : non-negative float, default=0.0
        Complexity parameter used for Minimal Cost-Complexity Pruning. The
        subtree with the largest cost complexity that is smaller than
        ``ccp_alpha`` will be chosen. By default, no pruning is performed. See
        :ref:`minimal_cost_complexity_pruning` for details.

        .. versionadded:: 0.22

    Attributes
    ----------
    max_features_ : int
        The inferred value of max_features.

    n_features_in_ : int
        Number of features seen during :term:`fit`.

    feature_names_in_ : ndarray of shape (`n_features_in_`,)
        Names of features seen during :term:`fit`. Defined only when `X`
        has feature names that are all strings.

    feature_importances_ : ndarray of shape (n_features,)
        Return impurity-based feature importances (the higher, the more
        important the feature).

        Warning: impurity-based feature importances can be misleading for
        high cardinality features (many unique values). See
        :func:`sklearn.inspection.permutation_importance` as an alternative.

    n_outputs_ : int
        The number of outputs when ``fit`` is performed.

    tree_ : Tree instance
        The underlying Tree object. Please refer to
        ``help(sklearn.tree._tree.Tree)`` for attributes of Tree object and
        :ref:`sphx_glr_auto_examples_tree_plot_unveil_tree_structure.py`
        for basic usage of these attributes.

    See Also
    --------
    ExtraTreeClassifier : An extremely randomized tree classifier.
    sklearn.ensemble.ExtraTreesClassifier : An extra-trees classifier.
    sklearn.ensemble.ExtraTreesRegressor : An extra-trees regressor.

    Notes
    -----
    The default values for the parameters controlling the size of the trees
    (e.g. ``max_depth``, ``min_samples_leaf``, etc.) lead to fully grown and
    unpruned trees which can potentially be very large on some data sets. To
    reduce memory consumption, the complexity and size of the trees should be
    controlled by setting those parameter values.

    References
    ----------

    .. [1] P. Geurts, D. Ernst., and L. Wehenkel, "Extremely randomized trees",
           Machine Learning, 63(1), 3-42, 2006.

    Examples
    --------
    >>> from sklearn.datasets import load_diabetes
    >>> from sklearn.model_selection import train_test_split
    >>> from sklearn.ensemble import BaggingRegressor
    >>> from sklearn.tree import ExtraTreeRegressor
    >>> X, y = load_diabetes(return_X_y=True)
    >>> X_train, X_test, y_train, y_test = train_test_split(
    ...     X, y, random_state=0)
    >>> extra_tree = ExtraTreeRegressor(random_state=0)
    >>> reg = BaggingRegressor(extra_tree, random_state=0).fit(
    ...     X_train, y_train)
    >>> reg.score(X_test, y_test)
    0.33...
    """

    def __init__(
        self,
        *,
        criterion="squared_error",
        splitter="random",
        max_depth=None,
        min_samples_split=2,
        min_samples_leaf=1,
        min_weight_fraction_leaf=0.0,
        max_features=1.0,
        random_state=None,
        min_impurity_decrease=0.0,
        max_leaf_nodes=None,
        ccp_alpha=0.0,
        # Semi-supervised parameters:
        supervision=0.5,
        ss_adapter=None,
        unsupervised_criterion="squared_error",
        update_supervision=None,
        preprocess_X_targets=None,
        _X_targets=None,
    ):
        super().__init__(
            splitter=splitter,
            max_depth=max_depth,
            min_samples_split=min_samples_split,
            min_samples_leaf=min_samples_leaf,
            min_weight_fraction_leaf=min_weight_fraction_leaf,
            max_features=max_features,
            max_leaf_nodes=max_leaf_nodes,
            min_impurity_decrease=min_impurity_decrease,
            random_state=random_state,
            ccp_alpha=ccp_alpha,
            # Semi-supervised parameters:
            supervision=supervision,
            ss_adapter=ss_adapter,
            criterion=criterion,
            unsupervised_criterion=unsupervised_criterion,
            update_supervision=update_supervision,
            preprocess_X_targets=preprocess_X_targets,
            _X_targets=_X_targets,
        )


# =============================================================================
# Bipartite semi-supervised trees
# =============================================================================


class BaseBipartiteDecisionTreeSS(
    BaseBipartiteDecisionTree,
    metaclass=ABCMeta,
):
    _parameter_constraints: dict = {
        **BaseBipartiteDecisionTree._parameter_constraints,
        **BaseDecisionTreeSS._parameter_constraints,
        "unsupervised_criterion_rows": [
            StrOptions(set(UNSUPERVISED_CRITERIA.keys())),
            Hidden(Criterion),
        ],
        "unsupervised_criterion_cols": [
            StrOptions(set(UNSUPERVISED_CRITERIA.keys())),
            Hidden(Criterion),
        ],
        "preprocess_X_targets": [callable, None],
    }
    _parameter_constraints.pop("unsupervised_criterion")

    @abstractmethod
    def __init__(
        self,
        *,
        criterion,
        splitter,
        max_depth,
        min_samples_split,
        min_samples_leaf,
        min_weight_fraction_leaf,
        max_features,
        max_leaf_nodes,
        random_state,
        min_impurity_decrease,
        class_weight=None,
        ccp_alpha=0.0,
        # Bipartite parameters:
        min_rows_split=1,  # Not 2, to still allow splitting on the other axis
        min_cols_split=1,
        min_rows_leaf=1,
        min_cols_leaf=1,
        min_row_weight_fraction_leaf=0.0,
        min_col_weight_fraction_leaf=0.0,
        max_row_features=None,
        max_col_features=None,
        bipartite_adapter="gmosa",
        prediction_weights=None,
        # Semi-supervised parameters:
        supervision=0.5,
        ss_adapter=None,
        unsupervised_criterion_rows="squared_error",
        unsupervised_criterion_cols="squared_error",
        update_supervision=None,
        axis_decision_only=False,
        preprocess_X_targets=None,
    ):
        self.supervision = supervision
        self.ss_adapter = ss_adapter
        self.unsupervised_criterion_rows = unsupervised_criterion_rows
        self.unsupervised_criterion_cols = unsupervised_criterion_cols
        self.update_supervision = update_supervision
        self.axis_decision_only = axis_decision_only
        self.preprocess_X_targets = preprocess_X_targets

        BaseBipartiteDecisionTree.__init__(
            self,
            criterion=criterion,
            splitter=splitter,
            max_depth=max_depth,
            min_samples_split=min_samples_split,
            min_samples_leaf=min_samples_leaf,
            min_weight_fraction_leaf=min_weight_fraction_leaf,
            max_features=max_features,
            max_leaf_nodes=max_leaf_nodes,
            random_state=random_state,
            min_impurity_decrease=min_impurity_decrease,
            class_weight=class_weight,
            ccp_alpha=ccp_alpha,
            min_rows_split=min_rows_split,
            min_cols_split=min_cols_split,
            min_rows_leaf=min_rows_leaf,
            min_cols_leaf=min_cols_leaf,
            min_row_weight_fraction_leaf=min_row_weight_fraction_leaf,
            min_col_weight_fraction_leaf=min_col_weight_fraction_leaf,
            max_row_features=max_row_features,
            max_col_features=max_col_features,
            bipartite_adapter=bipartite_adapter,
            prediction_weights=prediction_weights,
        )

    def _more_tags(self):
        # For cross-validation routines to split data correctly
        tags = {
            "pairwise_rows": (
                _is_pairwise_criterion(self.unsupervised_criterion_rows)
                or self.prediction_weights is not None
            ),
            "pairwise_cols": (
                _is_pairwise_criterion(self.unsupervised_criterion_cols)
                or self.prediction_weights is not None
            )
        }
        tags["pairwise"] = tags["pairwise_rows"] and tags["pairwise_cols"]
        return tags

    @_fit_context(prefer_skip_nested_validation=True)
    def fit(self, X, y, sample_weight=None, check_input=True, _X_targets=None):
        if check_input:
            _X_targets = self._check_X_targets(X, _X_targets)

        self._X_targets = _X_targets  # TODO: Keep it?

        return super().fit(X, y, sample_weight, check_input)

    def _check_X_targets(self, X, _X_targets=None):
        if _X_targets is None and self.preprocess_X_targets is not None:
            X = [self.preprocess_X_targets(Xi) for Xi in X]

        return _validate_bipartite_X_targets(
            X=X,
            X_targets=_X_targets,
            is_categoric_rows=_is_categoric_criterion(
                self.unsupervised_criterion_rows,
            ),
            is_categoric_cols=_is_categoric_criterion(
                self.unsupervised_criterion_cols,
            ),
            is_pairwise_rows=_is_pairwise_criterion(
                self.unsupervised_criterion_rows,
            ),
            is_pairwise_cols=_is_pairwise_criterion(
                self.unsupervised_criterion_cols,
            ),
        )

    def _make_splitter(
        self,
        *,
        X,
        n_outputs,
        n_samples,
        n_classes,
        min_samples_leaf,
        min_weight_leaf,
        ax_max_features,
        ax_min_samples_leaf,
        ax_min_weight_leaf,
        random_state,
    ):
        if isinstance(self.splitter, BipartiteSplitter):
            # Make a deepcopy in case the splitter has mutable attributes that
            # might be shared and modified concurrently during parallel fitting
            return deepcopy(self.splitter)

        # NOTE: It is possible to use diferent ss_adapter for rows and columns,
        #       but the user needs to pass an already built BipartiteSplitter instance
        #       if they want to do so.
        if isinstance(self.ss_adapter, _semisupervised_criterion.SSCompositeCriterion):
            ss_adapter = deepcopy(self.ss_adapter)
        elif isinstance(self.ss_adapter, str):
            ss_adapter = SS_CRITERIA[self.ss_adapter]
        elif self.ss_adapter is None:
            ss_adapter = None
        else:
            raise ValueError  # validate_params will not allow this to run

        if isinstance(self.unsupervised_criterion_rows, str):
            unsup_criterion_rows = \
                UNSUPERVISED_CRITERIA[self.unsupervised_criterion_rows]
        else:
            unsup_criterion_rows = deepcopy(self.unsupervised_criterion_rows)

        if isinstance(self.unsupervised_criterion_cols, str):
            unsup_criterion_cols = \
                UNSUPERVISED_CRITERIA[self.unsupervised_criterion_cols]
        else:
            unsup_criterion_cols = deepcopy(self.unsupervised_criterion_cols)

        bipartite_adapter = self.bipartite_adapter
        criterion = self.criterion

        if isinstance(criterion, str) and isinstance(bipartite_adapter, str):
            bipartite_adapter, criterion = _get_criterion_classes(
                adapter=bipartite_adapter,
                criterion=criterion,
                is_classification=is_classifier(self),
            )
        elif isinstance(criterion, str) or isinstance(bipartite_adapter, str):
            raise ValueError(
                "Either both or none of criterion and bipartite_adapter params"
                " must be strings."
            )
        else:  # Criterion instance or subclass
            criterion = deepcopy(criterion)

        splitter = self.splitter

        # User is able to specify a splitter for each axis
        if not isinstance(splitter, (tuple, list)):
            splitter = [splitter, splitter]
        for ax in range(2):
            if isinstance(splitter[ax], str):
                if issparse(X[ax]):
                    splitter[ax] = SPARSE_SPLITTERS[splitter[ax]]
                else:
                    splitter[ax] = DENSE_SPLITTERS[splitter[ax]]
            else:  # is a Splitter instance
                splitter[ax] = deepcopy(splitter[ax])

        splitter = _splitter_factory.make_bipartite_ss_splitter(
            splitters=splitter,
            supervised_criteria=criterion,
            unsupervised_criteria=[
                unsup_criterion_rows,
                unsup_criterion_cols,
            ],
            ss_criteria=ss_adapter,
            supervision=self.supervision,
            update_supervision=self.update_supervision,
            n_samples=n_samples,
            n_features=[
                self.n_row_features_in_,
                self.n_col_features_in_,
            ],
            n_classes=n_classes,
            n_outputs=n_outputs,
            max_features=ax_max_features,
            min_samples_leaf=min_samples_leaf,
            min_weight_leaf=min_weight_leaf,
            ax_min_samples_leaf=ax_min_samples_leaf,
            ax_min_weight_leaf=ax_min_weight_leaf,
            random_state=random_state,
            bipartite_criterion_class=bipartite_adapter,
            axis_decision_only=self.axis_decision_only,
        )

        if self._X_targets is None:
            raise RuntimeError("_X_targets was not set.")

        splitter.bipartite_criterion.set_X(
            self._X_targets[0],
            self._X_targets[1],
        )
        return splitter


class BipartiteDecisionTreeRegressorSS(
    BaseBipartiteDecisionTreeSS,
    BipartiteDecisionTreeRegressor,
):

    _parameter_constraints: dict = {
        **BaseBipartiteDecisionTreeSS._parameter_constraints,
        **BipartiteDecisionTreeRegressor._parameter_constraints,
    }
    _parameter_constraints.pop("splitter")

    def __init__(
        self,
        *,
        criterion="squared_error",
        splitter="best",
        max_depth=None,
        min_samples_split=2,
        min_samples_leaf=1,
        min_weight_fraction_leaf=0.0,
        max_features=None,
        random_state=None,
        max_leaf_nodes=None,
        min_impurity_decrease=0.0,
        ccp_alpha=0.0,
        # Bipartite parameters:
        min_rows_split=1,  # Not 2, to still allow splitting on the other axis
        min_cols_split=1,
        min_rows_leaf=1,
        min_cols_leaf=1,
        min_row_weight_fraction_leaf=0.0,
        min_col_weight_fraction_leaf=0.0,
        max_row_features=None,
        max_col_features=None,
        bipartite_adapter="gmosa",
        prediction_weights=None,
        # Semi-supervised parameters:
        supervision=0.5,
        ss_adapter=None,
        unsupervised_criterion_rows="squared_error",
        unsupervised_criterion_cols="squared_error",
        update_supervision=None,
        axis_decision_only=False,
        preprocess_X_targets=None,
    ):
        super().__init__(
            criterion=criterion,
            splitter=splitter,
            max_depth=max_depth,
            min_samples_split=min_samples_split,
            min_samples_leaf=min_samples_leaf,
            min_weight_fraction_leaf=min_weight_fraction_leaf,
            max_features=max_features,
            max_leaf_nodes=max_leaf_nodes,
            random_state=random_state,
            min_impurity_decrease=min_impurity_decrease,
            ccp_alpha=ccp_alpha,
            # Bipartite parameters:
            min_rows_split=min_rows_split,
            min_cols_split=min_cols_split,
            min_rows_leaf=min_rows_leaf,
            min_cols_leaf=min_cols_leaf,
            min_row_weight_fraction_leaf=min_row_weight_fraction_leaf,
            min_col_weight_fraction_leaf=min_col_weight_fraction_leaf,
            max_row_features=max_row_features,
            max_col_features=max_col_features,
            bipartite_adapter=bipartite_adapter,
            prediction_weights=prediction_weights,
            # Semi-supervised parameters:
            supervision=supervision,
            ss_adapter=ss_adapter,
            unsupervised_criterion_rows=unsupervised_criterion_rows,
            unsupervised_criterion_cols=unsupervised_criterion_cols,
            update_supervision=update_supervision,
            axis_decision_only=axis_decision_only,
            preprocess_X_targets=preprocess_X_targets,
        )


class BipartiteExtraTreeRegressorSS(
    BaseBipartiteDecisionTreeSS,
    BipartiteExtraTreeRegressor,
):

    _parameter_constraints: dict = {
        **BaseBipartiteDecisionTreeSS._parameter_constraints,
        **BipartiteExtraTreeRegressor._parameter_constraints,
    }
    _parameter_constraints.pop("splitter")

    def __init__(
        self,
        *,
        criterion="squared_error",
        splitter="random",
        max_depth=None,
        min_samples_split=2,
        min_samples_leaf=1,
        min_weight_fraction_leaf=0.0,
        max_features=1.0,
        random_state=None,
        max_leaf_nodes=None,
        min_impurity_decrease=0.0,
        ccp_alpha=0.0,
        # Bipartite parameters:
        min_rows_split=1,  # Not 2, to still allow splitting on the other axis
        min_cols_split=1,
        min_rows_leaf=1,
        min_cols_leaf=1,
        min_row_weight_fraction_leaf=0.0,
        min_col_weight_fraction_leaf=0.0,
        max_row_features=None,
        max_col_features=None,
        bipartite_adapter="gmosa",
        prediction_weights=None,
        # Semi-supervised parameters:
        supervision=0.5,
        ss_adapter=None,
        unsupervised_criterion_rows="squared_error",
        unsupervised_criterion_cols="squared_error",
        update_supervision=None,
        axis_decision_only=False,
        preprocess_X_targets=None,
    ):
        super().__init__(
            criterion=criterion,
            splitter=splitter,
            max_depth=max_depth,
            min_samples_split=min_samples_split,
            min_samples_leaf=min_samples_leaf,
            min_weight_fraction_leaf=min_weight_fraction_leaf,
            max_features=max_features,
            max_leaf_nodes=max_leaf_nodes,
            random_state=random_state,
            min_impurity_decrease=min_impurity_decrease,
            ccp_alpha=ccp_alpha,
            # Bipartite parameters:
            min_rows_split=min_rows_split,
            min_cols_split=min_cols_split,
            min_rows_leaf=min_rows_leaf,
            min_cols_leaf=min_cols_leaf,
            min_row_weight_fraction_leaf=min_row_weight_fraction_leaf,
            min_col_weight_fraction_leaf=min_col_weight_fraction_leaf,
            max_row_features=max_row_features,
            max_col_features=max_col_features,
            bipartite_adapter=bipartite_adapter,
            prediction_weights=prediction_weights,
            # Semi-supervised parameters:
            supervision=supervision,
            ss_adapter=ss_adapter,
            unsupervised_criterion_rows=unsupervised_criterion_rows,
            unsupervised_criterion_cols=unsupervised_criterion_cols,
            update_supervision=update_supervision,
            axis_decision_only=axis_decision_only,
            preprocess_X_targets=preprocess_X_targets,
        )
